# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

Current = 1  # type: Literal[1]
NewFirst = 2  # type: Literal[2]
NewSecond = 3  # type: Literal[3]
WipeCodeFirst = 4  # type: Literal[4]
WipeCodeSecond = 5  # type: Literal[5]



#!/usr/bin/env python3

import sys
import click

from trezorlib import firmware


@click.command()
@click.argument("filename", type=click.File("rb"))
@click.option("-o", "--output", type=click.File("w"), default="-")
def firmware_fingerprint(filename, output):
    """Display fingerprint of a firmware file."""
    data = filename.read()

    try:
        version, fw = firmware.parse(data)
    except Exception as e:
        click.echo(e, err=True)
        sys.exit(2)

    fingerprint = firmware.digest(version, fw).hex()
    click.echo(fingerprint, file=output)


if __name__ == "__main__":
    firmware_fingerprint()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .RipplePayment import RipplePayment

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class RippleSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 402

    def __init__(
        self,
        address_n: List[int] = None,
        fee: int = None,
        flags: int = None,
        sequence: int = None,
        last_ledger_sequence: int = None,
        payment: RipplePayment = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.fee = fee
        self.flags = flags
        self.sequence = sequence
        self.last_ledger_sequence = last_ledger_sequence
        self.payment = payment

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            2: ('fee', p.UVarintType, 0),
            3: ('flags', p.UVarintType, 0),
            4: ('sequence', p.UVarintType, 0),
            5: ('last_ledger_sequence', p.UVarintType, 0),
            6: ('payment', RipplePayment, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from decimal import Decimal

from . import exceptions, messages
from .tools import expect, normalize_nfc, session


def from_json(json_dict):
    def make_input(vin):
        i = messages.TxInputType()
        if "coinbase" in vin:
            i.prev_hash = b"\0" * 32
            i.prev_index = 0xFFFFFFFF  # signed int -1
            i.script_sig = bytes.fromhex(vin["coinbase"])
            i.sequence = vin["sequence"]

        else:
            i.prev_hash = bytes.fromhex(vin["txid"])
            i.prev_index = vin["vout"]
            i.script_sig = bytes.fromhex(vin["scriptSig"]["hex"])
            i.sequence = vin["sequence"]

        return i

    def make_bin_output(vout):
        o = messages.TxOutputBinType()
        o.amount = int(Decimal(vout["value"]) * (10 ** 8))
        o.script_pubkey = bytes.fromhex(vout["scriptPubKey"]["hex"])
        return o

    t = messages.TransactionType()
    t.version = json_dict["version"]
    t.lock_time = json_dict.get("locktime")
    t.inputs = [make_input(vin) for vin in json_dict["vin"]]
    t.bin_outputs = [make_bin_output(vout) for vout in json_dict["vout"]]
    return t


@expect(messages.PublicKey)
def get_public_node(
    client,
    n,
    ecdsa_curve_name=None,
    show_display=False,
    coin_name=None,
    script_type=messages.InputScriptType.SPENDADDRESS,
    ignore_xpub_magic=False,
):
    return client.call(
        messages.GetPublicKey(
            address_n=n,
            ecdsa_curve_name=ecdsa_curve_name,
            show_display=show_display,
            coin_name=coin_name,
            script_type=script_type,
            ignore_xpub_magic=ignore_xpub_magic,
        )
    )


@expect(messages.Address, field="address")
def get_address(
    client,
    coin_name,
    n,
    show_display=False,
    multisig=None,
    script_type=messages.InputScriptType.SPENDADDRESS,
):
    return client.call(
        messages.GetAddress(
            address_n=n,
            coin_name=coin_name,
            show_display=show_display,
            multisig=multisig,
            script_type=script_type,
        )
    )


@expect(messages.OwnershipId, field="ownership_id")
def get_ownership_id(
    client,
    coin_name,
    n,
    multisig=None,
    script_type=messages.InputScriptType.SPENDADDRESS,
):
    return client.call(
        messages.GetOwnershipId(
            address_n=n,
            coin_name=coin_name,
            multisig=multisig,
            script_type=script_type,
        )
    )


def get_ownership_proof(
    client,
    coin_name,
    n,
    multisig=None,
    script_type=messages.InputScriptType.SPENDADDRESS,
    user_confirmation=False,
    ownership_ids=None,
    commitment_data=None,
    preauthorized=False,
):
    if preauthorized:
        res = client.call(messages.DoPreauthorized())
        if not isinstance(res, messages.PreauthorizedRequest):
            raise exceptions.TrezorException("Unexpected message")

    res = client.call(
        messages.GetOwnershipProof(
            address_n=n,
            coin_name=coin_name,
            script_type=script_type,
            multisig=multisig,
            user_confirmation=user_confirmation,
            ownership_ids=ownership_ids,
            commitment_data=commitment_data,
        )
    )

    if not isinstance(res, messages.OwnershipProof):
        raise exceptions.TrezorException("Unexpected message")

    return res.ownership_proof, res.signature


@expect(messages.MessageSignature)
def sign_message(
    client, coin_name, n, message, script_type=messages.InputScriptType.SPENDADDRESS
):
    message = normalize_nfc(message)
    return client.call(
        messages.SignMessage(
            coin_name=coin_name, address_n=n, message=message, script_type=script_type
        )
    )


def verify_message(client, coin_name, address, signature, message):
    message = normalize_nfc(message)
    try:
        resp = client.call(
            messages.VerifyMessage(
                address=address,
                signature=signature,
                message=message,
                coin_name=coin_name,
            )
        )
    except exceptions.TrezorFailure:
        return False
    return isinstance(resp, messages.Success)


@session
def sign_tx(
    client,
    coin_name,
    inputs,
    outputs,
    details=None,
    prev_txes=None,
    preauthorized=False,
):
    this_tx = messages.TransactionType(inputs=inputs, outputs=outputs)

    if details is None:
        signtx = messages.SignTx()
    else:
        signtx = details

    signtx.coin_name = coin_name
    signtx.inputs_count = len(inputs)
    signtx.outputs_count = len(outputs)

    if preauthorized:
        res = client.call(messages.DoPreauthorized())
        if not isinstance(res, messages.PreauthorizedRequest):
            raise exceptions.TrezorException("Unexpected message")

    res = client.call(signtx)

    # Prepare structure for signatures
    signatures = [None] * len(inputs)
    serialized_tx = b""

    def copy_tx_meta(tx):
        tx_copy = messages.TransactionType(**tx)
        # clear fields
        tx_copy.inputs_cnt = len(tx.inputs)
        tx_copy.inputs = []
        tx_copy.outputs_cnt = len(tx.bin_outputs or tx.outputs)
        tx_copy.outputs = []
        tx_copy.bin_outputs = []
        tx_copy.extra_data_len = len(tx.extra_data or b"")
        tx_copy.extra_data = None
        return tx_copy

    R = messages.RequestType
    while isinstance(res, messages.TxRequest):
        # If there's some part of signed transaction, let's add it
        if res.serialized:
            if res.serialized.serialized_tx:
                serialized_tx += res.serialized.serialized_tx

            if res.serialized.signature_index is not None:
                idx = res.serialized.signature_index
                sig = res.serialized.signature
                if signatures[idx] is not None:
                    raise ValueError("Signature for index %d already filled" % idx)
                signatures[idx] = sig

        if res.request_type == R.TXFINISHED:
            break

        # Device asked for one more information, let's process it.
        if res.details.tx_hash is not None:
            current_tx = prev_txes[res.details.tx_hash]
        else:
            current_tx = this_tx

        if res.request_type == R.TXMETA:
            msg = copy_tx_meta(current_tx)
            res = client.call(messages.TxAck(tx=msg))

        elif res.request_type == R.TXINPUT:
            msg = messages.TransactionType()
            msg.inputs = [current_tx.inputs[res.details.request_index]]
            res = client.call(messages.TxAck(tx=msg))

        elif res.request_type == R.TXOUTPUT:
            msg = messages.TransactionType()
            if res.details.tx_hash:
                msg.bin_outputs = [current_tx.bin_outputs[res.details.request_index]]
            else:
                msg.outputs = [current_tx.outputs[res.details.request_index]]

            res = client.call(messages.TxAck(tx=msg))

        elif res.request_type == R.TXEXTRADATA:
            o, l = res.details.extra_data_offset, res.details.extra_data_len
            msg = messages.TransactionType()
            msg.extra_data = current_tx.extra_data[o : o + l]
            res = client.call(messages.TxAck(tx=msg))

    if not isinstance(res, messages.TxRequest):
        raise exceptions.TrezorException("Unexpected message")

    for i, sig in zip(inputs, signatures):
        if i.script_type != messages.InputScriptType.EXTERNAL and sig is None:
            raise exceptions.TrezorException("Some signatures are missing!")

    return signatures, serialized_tx


@expect(messages.Success, field="message")
def authorize_coinjoin(
    client,
    coordinator,
    max_total_fee,
    n,
    coin_name,
    fee_per_anonymity=None,
    script_type=messages.InputScriptType.SPENDADDRESS,
):
    return client.call(
        messages.AuthorizeCoinJoin(
            coordinator=coordinator,
            max_total_fee=max_total_fee,
            address_n=n,
            coin_name=coin_name,
            fee_per_anonymity=fee_per_anonymity,
            script_type=script_type,
        )
    )



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

NORMAL = 0  # type: Literal[0]
BOLD = 1  # type: Literal[1]
MONO = 2  # type: Literal[2]
BR = 4  # type: Literal[4]
BR_HALF = 5  # type: Literal[5]
SET_COLOR = 6  # type: Literal[6]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroRingCtSig(p.MessageType):

    def __init__(
        self,
        txn_fee: int = None,
        message: bytes = None,
        rv_type: int = None,
    ) -> None:
        self.txn_fee = txn_fee
        self.message = message
        self.rv_type = rv_type

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('txn_fee', p.UVarintType, 0),
            2: ('message', p.BytesType, 0),
            3: ('rv_type', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

TXINPUT = 0  # type: Literal[0]
TXOUTPUT = 1  # type: Literal[1]
TXMETA = 2  # type: Literal[2]
TXFINISHED = 3  # type: Literal[3]
TXEXTRADATA = 4  # type: Literal[4]



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

TIF_UNKNOWN = 0  # type: Literal[0]
GTE = 1  # type: Literal[1]
TIF_RESERVED = 2  # type: Literal[2]
IOC = 3  # type: Literal[3]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class TxRequestDetailsType(p.MessageType):

    def __init__(
        self,
        request_index: int = None,
        tx_hash: bytes = None,
        extra_data_len: int = None,
        extra_data_offset: int = None,
    ) -> None:
        self.request_index = request_index
        self.tx_hash = tx_hash
        self.extra_data_len = extra_data_len
        self.extra_data_offset = extra_data_offset

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('request_index', p.UVarintType, 0),
            2: ('tx_hash', p.BytesType, 0),
            3: ('extra_data_len', p.UVarintType, 0),
            4: ('extra_data_offset', p.UVarintType, 0),
        }



#!/usr/bin/env python3
import hashlib

import mnemonic

__doc__ = '''
    Use this script to cross-check that Trezor generated valid
    mnemonic sentence for given internal (Trezor-generated)
    and external (computer-generated) entropy.

    Keep in mind that you're entering secret information to this script.
    Leaking of these information may lead to stealing your bitcoins
    from your wallet! We strongly recommend to run this script only on
    highly secured computer (ideally live linux distribution
    without an internet connection).
'''


def generate_entropy(strength, internal_entropy, external_entropy):
    '''
    strength - length of produced seed. One of 128, 192, 256
    random - binary stream of random data from external HRNG
    '''
    if strength not in (128, 192, 256):
        raise ValueError("Invalid strength")

    if not internal_entropy:
        raise ValueError("Internal entropy is not provided")

    if len(internal_entropy) < 32:
        raise ValueError("Internal entropy too short")

    if not external_entropy:
        raise ValueError("External entropy is not provided")

    if len(external_entropy) < 32:
        raise ValueError("External entropy too short")

    entropy = hashlib.sha256(internal_entropy + external_entropy).digest()
    entropy_stripped = entropy[:strength // 8]

    if len(entropy_stripped) * 8 != strength:
        raise ValueError("Entropy length mismatch")

    return entropy_stripped


def main():
    print(__doc__)

    comp = bytes.fromhex(input("Please enter computer-generated entropy (in hex): ").strip())
    trzr = bytes.fromhex(input("Please enter Trezor-generated entropy (in hex): ").strip())
    word_count = int(input("How many words your mnemonic has? "))

    strength = word_count * 32 // 3

    entropy = generate_entropy(strength, trzr, comp)

    words = mnemonic.Mnemonic('english').to_mnemonic(entropy)
    if not mnemonic.Mnemonic('english').check(words):
        print("Mnemonic is invalid")
        return

    if len(words.split(' ')) != word_count:
        print("Mnemonic length mismatch!")
        return

    print("Generated mnemonic is:", words)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
from trezorlib.debuglink import DebugLink
from trezorlib.transport import enumerate_devices
import sys

# usage examples
# read entire bootloader: ./mem_read.py 8000000 8000
# read initial stack pointer: ./mem_read.py 8000000 4
# an entire bootloader can be later disassembled with:
# arm-none-eabi-objdump -D -b binary -m arm -M force-thumb memory.dat
# note that in order for this to work, your trezor device must
# be running a firmware that was built with debug link enabled


def find_debug():
    for device in enumerate_devices():
        try:
            debug_transport = device.find_debug()
            debug = DebugLink(debug_transport, auto_interact=False)
            debug.open()
            return debug
        except Exception:
            continue
    else:
        print("No suitable Trezor device found")
        sys.exit(1)


def main():
    debug = find_debug()

    arg1 = int(sys.argv[1], 16)
    arg2 = int(sys.argv[2], 16)
    step = 0x400 if arg2 >= 0x400 else arg2

    f = open("memory.dat", "wb")

    for addr in range(arg1, arg1 + arg2, step):
        mem = debug.memory_read(addr, step)
        f.write(mem)

    f.close()


if __name__ == "__main__":
    main()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeTezosContractType = Literal[0, 1]
    except ImportError:
        pass


class TezosContractID(p.MessageType):

    def __init__(
        self,
        tag: EnumTypeTezosContractType = None,
        hash: bytes = None,
    ) -> None:
        self.tag = tag
        self.hash = hash

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('tag', p.EnumType("TezosContractType", (0, 1)), 0),
            2: ('hash', p.BytesType, 0),
        }



#!/usr/bin/env python3
# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import json

import click
import decimal
import requests

from trezorlib import btc, messages, tools
from trezorlib.cli import ChoiceType
from trezorlib.cli.btc import INPUT_SCRIPTS, OUTPUT_SCRIPTS
from trezorlib.protobuf import to_dict


def echo(*args, **kwargs):
    return click.echo(*args, err=True, **kwargs)


def prompt(*args, **kwargs):
    return click.prompt(*args, err=True, **kwargs)


def _default_script_type(address_n, script_types):
    script_type = "address"

    if address_n is None:
        pass
    elif address_n[0] == tools.H_(49):
        script_type = "p2shsegwit"
    elif address_n[0] == tools.H_(84):
        script_type = "segwit"

    return script_type
    # return script_types[script_type]


def parse_vin(s):
    txid, vout = s.split(":")
    return bytes.fromhex(txid), int(vout)


def _get_inputs_interactive(blockbook_url):
    inputs = []
    txes = {}
    while True:
        echo()
        prev = prompt(
            "Previous output to spend (txid:vout)", type=parse_vin, default=""
        )
        if not prev:
            break
        prev_hash, prev_index = prev
        address_n = prompt("BIP-32 path to derive the key", type=tools.parse_path)

        txhash = prev_hash.hex()
        tx_url = blockbook_url + txhash
        r = requests.get(tx_url)
        if not r.ok:
            raise click.ClickException(f"Failed to fetch URL: {tx_url}")

        tx_json = r.json(parse_float=decimal.Decimal)
        if "error" in tx_json:
            raise click.ClickException(f"Transaction not found: {txhash}")

        tx = btc.from_json(tx_json)
        txes[txhash] = tx
        amount = tx.bin_outputs[prev_index].amount
        echo("Input amount: {}".format(amount))

        sequence = prompt(
            "Sequence Number to use (RBF opt-in enabled by default)",
            type=int,
            default=0xFFFFFFFD,
        )
        script_type = prompt(
            "Input type",
            type=ChoiceType(INPUT_SCRIPTS),
            default=_default_script_type(address_n, INPUT_SCRIPTS),
        )
        if isinstance(script_type, str):
            script_type = INPUT_SCRIPTS[script_type]

        new_input = messages.TxInputType(
            address_n=address_n,
            prev_hash=prev_hash,
            prev_index=prev_index,
            amount=amount,
            script_type=script_type,
            sequence=sequence,
        )

        inputs.append(new_input)

    return inputs, txes


def _get_outputs_interactive():
    outputs = []
    while True:
        echo()
        address = prompt("Output address (for non-change output)", default="")
        if address:
            address_n = None
            script_type = messages.OutputScriptType.PAYTOADDRESS
        else:
            address = None
            address_n = prompt(
                "BIP-32 path (for change output)", type=tools.parse_path, default=""
            )
            if not address_n:
                break
            script_type = prompt(
                "Output type",
                type=ChoiceType(OUTPUT_SCRIPTS),
                default=_default_script_type(address_n, OUTPUT_SCRIPTS),
            )
            if isinstance(script_type, str):
                script_type = OUTPUT_SCRIPTS[script_type]

        amount = prompt("Amount to spend (satoshis)", type=int)

        outputs.append(
            messages.TxOutputType(
                address_n=address_n,
                address=address,
                amount=amount,
                script_type=script_type,
            )
        )

    return outputs


@click.command()
def sign_interactive():
    coin = prompt("Coin name", default="Bitcoin")
    blockbook_host = prompt("Blockbook server", default="btc1.trezor.io")

    if not requests.get(f"https://{blockbook_host}/api").ok:
        raise click.ClickException("Could not connect to blockbook")

    blockbook_url = f"https://{blockbook_host}/api/tx-specific/"

    inputs, txes = _get_inputs_interactive(blockbook_url)
    outputs = _get_outputs_interactive()

    signtx = messages.SignTx()
    signtx.version = prompt("Transaction version", type=int, default=2)
    signtx.lock_time = prompt("Transaction locktime", type=int, default=0)

    result = {
        "coin_name": coin,
        "inputs": [to_dict(i, hexlify_bytes=True) for i in inputs],
        "outputs": [to_dict(o, hexlify_bytes=True) for o in outputs],
        "details": to_dict(signtx, hexlify_bytes=True),
        "prev_txes": {
            txhash: to_dict(txdata, hexlify_bytes=True)
            for txhash, txdata in txes.items()
        },
    }

    print(json.dumps(result, sort_keys=True, indent=2))


if __name__ == "__main__":
    sign_interactive()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeCardanoCertificateType = Literal[0, 1, 2]
    except ImportError:
        pass


class CardanoTxCertificateType(p.MessageType):

    def __init__(
        self,
        type: EnumTypeCardanoCertificateType = None,
        path: List[int] = None,
        pool: bytes = None,
    ) -> None:
        self.type = type
        self.path = path if path is not None else []
        self.pool = pool

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('type', p.EnumType("CardanoCertificateType", (0, 1, 2)), 0),
            2: ('path', p.UVarintType, p.FLAG_REPEATED),
            3: ('pool', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroTransactionRsigData(p.MessageType):

    def __init__(
        self,
        rsig_type: int = None,
        offload_type: int = None,
        grouping: List[int] = None,
        mask: bytes = None,
        rsig: bytes = None,
        rsig_parts: List[bytes] = None,
        bp_version: int = None,
    ) -> None:
        self.rsig_type = rsig_type
        self.offload_type = offload_type
        self.grouping = grouping if grouping is not None else []
        self.mask = mask
        self.rsig = rsig
        self.rsig_parts = rsig_parts if rsig_parts is not None else []
        self.bp_version = bp_version

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('rsig_type', p.UVarintType, 0),
            2: ('offload_type', p.UVarintType, 0),
            3: ('grouping', p.UVarintType, p.FLAG_REPEATED),
            4: ('mask', p.BytesType, 0),
            5: ('rsig', p.BytesType, 0),
            6: ('rsig_parts', p.BytesType, p.FLAG_REPEATED),
            7: ('bp_version', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

STAKE_REGISTRATION = 0  # type: Literal[0]
STAKE_DEREGISTRATION = 1  # type: Literal[1]
STAKE_DELEGATION = 2  # type: Literal[2]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .StellarAssetType import StellarAssetType

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class StellarChangeTrustOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 216

    def __init__(
        self,
        source_account: str = None,
        asset: StellarAssetType = None,
        limit: int = None,
    ) -> None:
        self.source_account = source_account
        self.asset = asset
        self.limit = limit

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('source_account', p.UnicodeType, 0),
            2: ('asset', StellarAssetType, 0),
            3: ('limit', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class EosTxHeader(p.MessageType):

    def __init__(
        self,
        expiration: int = None,
        ref_block_num: int = None,
        ref_block_prefix: int = None,
        max_net_usage_words: int = None,
        max_cpu_usage_ms: int = None,
        delay_sec: int = None,
    ) -> None:
        self.expiration = expiration
        self.ref_block_num = ref_block_num
        self.ref_block_prefix = ref_block_prefix
        self.max_net_usage_words = max_net_usage_words
        self.max_cpu_usage_ms = max_cpu_usage_ms
        self.delay_sec = delay_sec

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('expiration', p.UVarintType, 0),  # required
            2: ('ref_block_num', p.UVarintType, 0),  # required
            3: ('ref_block_prefix', p.UVarintType, 0),  # required
            4: ('max_net_usage_words', p.UVarintType, 0),  # required
            5: ('max_cpu_usage_ms', p.UVarintType, 0),  # required
            6: ('delay_sec', p.UVarintType, 0),  # required
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class CardanoBlockchainPointerType(p.MessageType):

    def __init__(
        self,
        block_index: int = None,
        tx_index: int = None,
        certificate_index: int = None,
    ) -> None:
        self.block_index = block_index
        self.tx_index = tx_index
        self.certificate_index = certificate_index

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('block_index', p.UVarintType, 0),
            2: ('tx_index', p.UVarintType, 0),
            3: ('certificate_index', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroSubAddressIndicesList import MoneroSubAddressIndicesList

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroKeyImageExportInitRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 530

    def __init__(
        self,
        num: int = None,
        hash: bytes = None,
        address_n: List[int] = None,
        network_type: int = None,
        subs: List[MoneroSubAddressIndicesList] = None,
    ) -> None:
        self.num = num
        self.hash = hash
        self.address_n = address_n if address_n is not None else []
        self.network_type = network_type
        self.subs = subs if subs is not None else []

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('num', p.UVarintType, 0),
            2: ('hash', p.BytesType, 0),
            3: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            4: ('network_type', p.UVarintType, 0),
            5: ('subs', MoneroSubAddressIndicesList, p.FLAG_REPEATED),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

SPENDADDRESS = 0  # type: Literal[0]
SPENDMULTISIG = 1  # type: Literal[1]
EXTERNAL = 2  # type: Literal[2]
SPENDWITNESS = 3  # type: Literal[3]
SPENDP2SHWITNESS = 4  # type: Literal[4]



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import logging
from collections import namedtuple
from copy import deepcopy

from mnemonic import Mnemonic

from . import mapping, messages, protobuf
from .client import TrezorClient
from .exceptions import TrezorFailure
from .log import DUMP_BYTES
from .tools import expect

EXPECTED_RESPONSES_CONTEXT_LINES = 3

LayoutLines = namedtuple("LayoutLines", "lines text")

LOG = logging.getLogger(__name__)


def layout_lines(lines):
    return LayoutLines(lines, " ".join(lines))


class DebugLink:
    def __init__(self, transport, auto_interact=True):
        self.transport = transport
        self.allow_interactions = auto_interact

    def open(self):
        self.transport.begin_session()

    def close(self):
        self.transport.end_session()

    def _call(self, msg, nowait=False):
        LOG.debug(
            "sending message: {}".format(msg.__class__.__name__),
            extra={"protobuf": msg},
        )
        msg_type, msg_bytes = mapping.encode(msg)
        LOG.log(
            DUMP_BYTES,
            "encoded as type {} ({} bytes): {}".format(
                msg_type, len(msg_bytes), msg_bytes.hex()
            ),
        )
        self.transport.write(msg_type, msg_bytes)
        if nowait:
            return None

        ret_type, ret_bytes = self.transport.read()
        LOG.log(
            DUMP_BYTES,
            "received type {} ({} bytes): {}".format(
                msg_type, len(msg_bytes), msg_bytes.hex()
            ),
        )
        msg = mapping.decode(ret_type, ret_bytes)
        LOG.debug(
            "received message: {}".format(msg.__class__.__name__),
            extra={"protobuf": msg},
        )
        return msg

    def state(self):
        return self._call(messages.DebugLinkGetState())

    def read_layout(self):
        return layout_lines(self.state().layout_lines)

    def wait_layout(self):
        obj = self._call(messages.DebugLinkGetState(wait_layout=True))
        if isinstance(obj, messages.Failure):
            raise TrezorFailure(obj)
        return layout_lines(obj.layout_lines)

    def watch_layout(self, watch: bool) -> None:
        """Enable or disable watching layouts.
        If disabled, wait_layout will not work.

        The message is missing on T1. Use `TrezorClientDebugLink.watch_layout` for
        cross-version compatibility.
        """
        self._call(messages.DebugLinkWatchLayout(watch=watch))

    def encode_pin(self, pin, matrix=None):
        """Transform correct PIN according to the displayed matrix."""
        if matrix is None:
            matrix = self.state().matrix
            if matrix is None:
                # we are on trezor-core
                return pin

        return "".join([str(matrix.index(p) + 1) for p in pin])

    def read_recovery_word(self):
        state = self.state()
        return (state.recovery_fake_word, state.recovery_word_pos)

    def read_reset_word(self):
        state = self._call(messages.DebugLinkGetState(wait_word_list=True))
        return state.reset_word

    def read_reset_word_pos(self):
        state = self._call(messages.DebugLinkGetState(wait_word_pos=True))
        return state.reset_word_pos

    def input(self, word=None, button=None, swipe=None, x=None, y=None, wait=False):
        if not self.allow_interactions:
            return

        args = sum(a is not None for a in (word, button, swipe, x))
        if args != 1:
            raise ValueError("Invalid input - must use one of word, button, swipe")

        decision = messages.DebugLinkDecision(
            yes_no=button, swipe=swipe, input=word, x=x, y=y, wait=wait
        )
        ret = self._call(decision, nowait=not wait)
        if ret is not None:
            return layout_lines(ret.lines)

    def click(self, click, wait=False):
        x, y = click
        return self.input(x=x, y=y, wait=wait)

    def press_yes(self):
        self.input(button=True)

    def press_no(self):
        self.input(button=False)

    def swipe_up(self):
        self.input(swipe=messages.DebugSwipeDirection.UP)

    def swipe_down(self):
        self.input(swipe=messages.DebugSwipeDirection.DOWN)

    def swipe_right(self):
        self.input(swipe=messages.DebugSwipeDirection.RIGHT)

    def swipe_left(self):
        self.input(swipe=messages.DebugSwipeDirection.LEFT)

    def stop(self):
        self._call(messages.DebugLinkStop(), nowait=True)

    def reseed(self, value):
        self._call(messages.DebugLinkReseedRandom(value=value))

    def start_recording(self, directory):
        self._call(messages.DebugLinkRecordScreen(target_directory=directory))

    def stop_recording(self):
        self._call(messages.DebugLinkRecordScreen(target_directory=None))

    @expect(messages.DebugLinkMemory, field="memory")
    def memory_read(self, address, length):
        return self._call(messages.DebugLinkMemoryRead(address=address, length=length))

    def memory_write(self, address, memory, flash=False):
        self._call(
            messages.DebugLinkMemoryWrite(address=address, memory=memory, flash=flash),
            nowait=True,
        )

    def flash_erase(self, sector):
        self._call(messages.DebugLinkFlashErase(sector=sector), nowait=True)

    @expect(messages.Success)
    def erase_sd_card(self, format=True):
        return self._call(messages.DebugLinkEraseSdCard(format=format))


class NullDebugLink(DebugLink):
    def __init__(self):
        super().__init__(None)

    def open(self):
        pass

    def close(self):
        pass

    def _call(self, msg, nowait=False):
        if not nowait:
            if isinstance(msg, messages.DebugLinkGetState):
                return messages.DebugLinkState()
            else:
                raise RuntimeError("unexpected call to a fake debuglink")


class DebugUI:
    INPUT_FLOW_DONE = object()

    def __init__(self, debuglink: DebugLink):
        self.debuglink = debuglink
        self.clear()

    def clear(self):
        self.pins = None
        self.passphrase = ""
        self.input_flow = None

    def button_request(self, code):
        if self.input_flow is None:
            if code == messages.ButtonRequestType.PinEntry:
                self.debuglink.input(self.get_pin())
            else:
                self.debuglink.press_yes()
        elif self.input_flow is self.INPUT_FLOW_DONE:
            raise AssertionError("input flow ended prematurely")
        else:
            try:
                self.input_flow.send(code)
            except StopIteration:
                self.input_flow = self.INPUT_FLOW_DONE

    def get_pin(self, code=None):
        if self.pins is None:
            raise RuntimeError("PIN requested but no sequence was configured")

        try:
            return self.debuglink.encode_pin(next(self.pins))
        except StopIteration:
            raise AssertionError("PIN sequence ended prematurely")

    def get_passphrase(self, available_on_device):
        return self.passphrase


class TrezorClientDebugLink(TrezorClient):
    # This class implements automatic responses
    # and other functionality for unit tests
    # for various callbacks, created in order
    # to automatically pass unit tests.
    #
    # This mixing should be used only for purposes
    # of unit testing, because it will fail to work
    # without special DebugLink interface provided
    # by the device.

    def __init__(self, transport, auto_interact=True):
        try:
            debug_transport = transport.find_debug()
            self.debug = DebugLink(debug_transport, auto_interact)
            # try to open debuglink, see if it works
            self.debug.open()
            self.debug.close()
        except Exception:
            if not auto_interact:
                self.debug = NullDebugLink()
            else:
                raise

        self.ui = DebugUI(self.debug)

        self.in_with_statement = 0
        self.screenshot_id = 0

        self.filters = {}

        # Do not expect any specific response from device
        self.expected_responses = None
        self.current_response = None

        super().__init__(transport, ui=self.ui)

    def open(self):
        super().open()
        if self.session_counter == 1:
            self.debug.open()

    def close(self):
        if self.session_counter == 1:
            self.debug.close()
        super().close()

    def set_filter(self, message_type, callback):
        """Configure a filter function for a specified message type.

        The `callback` must be a function that accepts a protobuf message, and returns
        a (possibly modified) protobuf message of the same type. Whenever a message
        is sent or received that matches `message_type`, `callback` is invoked on the
        message and its result is substituted for the original.

        Useful for test scenarios with an active malicious actor on the wire.
        """
        self.filters[message_type] = callback

    def _filter_message(self, msg):
        message_type = msg.__class__
        callback = self.filters.get(message_type)
        if callable(callback):
            return callback(deepcopy(msg))
        else:
            return msg

    def set_input_flow(self, input_flow):
        """Configure a sequence of input events for the current with-block.

        The `input_flow` must be a generator function. A `yield` statement in the
        input flow function waits for a ButtonRequest from the device, and returns
        its code.

        Example usage:

        >>> def input_flow():
        >>>     # wait for first button prompt
        >>>     code = yield
        >>>     assert code == ButtonRequestType.Other
        >>>     # press No
        >>>     client.debug.press_no()
        >>>
        >>>     # wait for second button prompt
        >>>     yield
        >>>     # press Yes
        >>>     client.debug.press_yes()
        >>>
        >>> with client:
        >>>     client.set_input_flow(input_flow)
        >>>     some_call(client)
        """
        if not self.in_with_statement:
            raise RuntimeError("Must be called inside 'with' statement")

        if callable(input_flow):
            input_flow = input_flow()
        if not hasattr(input_flow, "send"):
            raise RuntimeError("input_flow should be a generator function")
        self.ui.input_flow = input_flow
        input_flow.send(None)  # start the generator

    def watch_layout(self, watch: bool = True) -> None:
        """Enable or disable watching layout changes.

        Since trezor-core v2.3.2, it is necessary to call `watch_layout()` before
        using `debug.wait_layout()`, otherwise layout changes are not reported.
        """
        if self.version >= (2, 3, 2):
            # version check is necessary because otherwise we cannot reliably detect
            # whether and where to wait for reply:
            # - T1 reports unknown debuglink messages on the wirelink
            # - TT < 2.3.0 does not reply to unknown debuglink messages due to a bug
            self.debug.watch_layout(watch)

    def __enter__(self):
        # For usage in with/expected_responses
        self.in_with_statement += 1
        return self

    def __exit__(self, _type, value, traceback):
        self.in_with_statement -= 1

        # Clear input flow.
        try:
            if _type is not None:
                # Another exception raised
                return False

            if self.expected_responses is None:
                # no need to check anything else
                return False

            # Evaluate missed responses in 'with' statement
            if self.current_response < len(self.expected_responses):
                self._raise_unexpected_response(None)

        finally:
            # Cleanup
            self.expected_responses = None
            self.current_response = None
            self.ui.clear()
            self.watch_layout(False)

        return False

    def set_expected_responses(self, expected):
        """Set a sequence of expected responses to client calls.

        Within a given with-block, the list of received responses from device must
        match the list of expected responses, otherwise an AssertionError is raised.

        If an expected response is given a field value other than None, that field value
        must exactly match the received field value. If a given field is None
        (or unspecified) in the expected response, the received field value is not
        checked.

        Each expected response can also be a tuple (bool, message). In that case, the
        expected response is only evaluated if the first field is True.
        This is useful for differentiating sequences between Trezor models:

        >>> trezor_one = client.features.model == "1"
        >>> client.set_expected_responses([
        >>>     messages.ButtonRequest(code=ConfirmOutput),
        >>>     (trezor_one, messages.ButtonRequest(code=ConfirmOutput)),
        >>>     messages.Success(),
        >>> ])
        """
        if not self.in_with_statement:
            raise RuntimeError("Must be called inside 'with' statement")

        # make sure all items are (bool, message) tuples
        expected_with_validity = [
            e if isinstance(e, tuple) else (True, e) for e in expected
        ]

        # only apply those items that are (True, message)
        self.expected_responses = [
            expected for valid, expected in expected_with_validity if valid
        ]

        self.current_response = 0

    def use_pin_sequence(self, pins):
        """Respond to PIN prompts from device with the provided PINs.
        The sequence must be at least as long as the expected number of PIN prompts.
        """
        self.ui.pins = iter(pins)

    def use_passphrase(self, passphrase):
        """Respond to passphrase prompts from device with the provided passphrase."""
        self.ui.passphrase = Mnemonic.normalize_string(passphrase)

    def use_mnemonic(self, mnemonic):
        """Use the provided mnemonic to respond to device.
        Only applies to T1, where device prompts the host for mnemonic words."""
        self.mnemonic = Mnemonic.normalize_string(mnemonic).split(" ")

    def _raw_read(self):
        __tracebackhide__ = True  # for pytest # pylint: disable=W0612

        resp = super()._raw_read()
        resp = self._filter_message(resp)
        self._check_request(resp)
        return resp

    def _raw_write(self, msg):
        return super()._raw_write(self._filter_message(msg))

    def _raise_unexpected_response(self, msg):
        __tracebackhide__ = True  # for pytest # pylint: disable=W0612

        start_at = max(self.current_response - EXPECTED_RESPONSES_CONTEXT_LINES, 0)
        stop_at = min(
            self.current_response + EXPECTED_RESPONSES_CONTEXT_LINES + 1,
            len(self.expected_responses),
        )
        output = []
        output.append("Expected responses:")
        if start_at > 0:
            output.append("    (...{} previous responses omitted)".format(start_at))
        for i in range(start_at, stop_at):
            exp = self.expected_responses[i]
            prefix = "    " if i != self.current_response else ">>> "
            set_fields = {
                key: value
                for key, value in exp.__dict__.items()
                if value is not None and value != []
            }
            oneline_str = ", ".join("{}={!r}".format(*i) for i in set_fields.items())
            if len(oneline_str) < 60:
                output.append(
                    "{}{}({})".format(prefix, exp.__class__.__name__, oneline_str)
                )
            else:
                item = []
                item.append("{}{}(".format(prefix, exp.__class__.__name__))
                for key, value in set_fields.items():
                    item.append("{}    {}={!r}".format(prefix, key, value))
                item.append("{})".format(prefix))
                output.append("\n".join(item))
        if stop_at < len(self.expected_responses):
            omitted = len(self.expected_responses) - stop_at
            output.append("    (...{} following responses omitted)".format(omitted))

        output.append("")
        if msg is not None:
            output.append("Actually received:")
            output.append(protobuf.format_message(msg))
        else:
            output.append("This message was never received.")
        raise AssertionError("\n".join(output))

    def _check_request(self, msg):
        __tracebackhide__ = True  # for pytest # pylint: disable=W0612
        if self.expected_responses is None:
            return

        if self.current_response >= len(self.expected_responses):
            raise AssertionError(
                "No more messages were expected, but we got:\n"
                + protobuf.format_message(msg)
            )

        expected = self.expected_responses[self.current_response]

        if msg.__class__ != expected.__class__:
            self._raise_unexpected_response(msg)

        for field, value in expected.__dict__.items():
            if value is None or value == []:
                continue
            if getattr(msg, field) != value:
                self._raise_unexpected_response(msg)

        self.current_response += 1

    def mnemonic_callback(self, _):
        word, pos = self.debug.read_recovery_word()
        if word != "":
            return word
        if pos != 0:
            return self.mnemonic[pos - 1]

        raise RuntimeError("Unexpected call")


@expect(messages.Success, field="message")
def load_device(
    client,
    mnemonic,
    pin,
    passphrase_protection,
    label,
    language="en-US",
    skip_checksum=False,
    needs_backup=False,
    no_backup=False,
):
    if not isinstance(mnemonic, (list, tuple)):
        mnemonic = [mnemonic]

    mnemonics = [Mnemonic.normalize_string(m) for m in mnemonic]

    if client.features.initialized:
        raise RuntimeError(
            "Device is initialized already. Call device.wipe() and try again."
        )

    resp = client.call(
        messages.LoadDevice(
            mnemonics=mnemonics,
            pin=pin,
            passphrase_protection=passphrase_protection,
            language=language,
            label=label,
            skip_checksum=skip_checksum,
            needs_backup=needs_backup,
            no_backup=no_backup,
        )
    )
    client.init_device()
    return resp


# keep the old name for compatibility
load_device_by_mnemonic = load_device


@expect(messages.Success, field="message")
def self_test(client):
    if client.features.bootloader_mode is not True:
        raise RuntimeError("Device must be in bootloader mode")

    return client.call(
        messages.SelfTest(
            payload=b"\x00\xFF\x55\xAA\x66\x99\x33\xCCABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!\x00\xFF\x55\xAA\x66\x99\x33\xCC"
        )
    )


@expect(messages.Success, field="message")
def show_text(client, header_text, body_text, icon=None, icon_color=None):
    body_text = [
        messages.DebugLinkShowTextItem(style=style, content=content)
        for style, content in body_text
    ]
    msg = messages.DebugLinkShowText(
        header_text=header_text,
        body_text=body_text,
        header_icon=icon,
        icon_color=icon_color,
    )
    return client.call(msg)



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from . import messages
from .tools import expect


@expect(messages.WebAuthnCredentials, field="credentials")
def list_credentials(client):
    return client.call(messages.WebAuthnListResidentCredentials())


@expect(messages.Success, field="message")
def add_credential(client, credential_id):
    return client.call(messages.WebAuthnAddResidentCredential(credential_id))


@expect(messages.Success, field="message")
def remove_credential(client, index):
    return client.call(messages.WebAuthnRemoveResidentCredential(index))


@expect(messages.Success, field="message")
def set_counter(client, u2f_counter):
    return client.call(messages.SetU2FCounter(u2f_counter=u2f_counter))


@expect(messages.NextU2FCounter, field="u2f_counter")
def get_next_counter(client):
    return client.call(messages.GetNextU2FCounter())



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

Bip39 = 0  # type: Literal[0]
Slip39_Basic = 1  # type: Literal[1]
Slip39_Advanced = 2  # type: Literal[2]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionDestinationEntry import MoneroTransactionDestinationEntry
from .MoneroTransactionRsigData import MoneroTransactionRsigData

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroTransactionData(p.MessageType):

    def __init__(
        self,
        version: int = None,
        payment_id: bytes = None,
        unlock_time: int = None,
        outputs: List[MoneroTransactionDestinationEntry] = None,
        change_dts: MoneroTransactionDestinationEntry = None,
        num_inputs: int = None,
        mixin: int = None,
        fee: int = None,
        account: int = None,
        minor_indices: List[int] = None,
        rsig_data: MoneroTransactionRsigData = None,
        integrated_indices: List[int] = None,
        client_version: int = None,
        hard_fork: int = None,
        monero_version: bytes = None,
    ) -> None:
        self.version = version
        self.payment_id = payment_id
        self.unlock_time = unlock_time
        self.outputs = outputs if outputs is not None else []
        self.change_dts = change_dts
        self.num_inputs = num_inputs
        self.mixin = mixin
        self.fee = fee
        self.account = account
        self.minor_indices = minor_indices if minor_indices is not None else []
        self.rsig_data = rsig_data
        self.integrated_indices = integrated_indices if integrated_indices is not None else []
        self.client_version = client_version
        self.hard_fork = hard_fork
        self.monero_version = monero_version

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('version', p.UVarintType, 0),
            2: ('payment_id', p.BytesType, 0),
            3: ('unlock_time', p.UVarintType, 0),
            4: ('outputs', MoneroTransactionDestinationEntry, p.FLAG_REPEATED),
            5: ('change_dts', MoneroTransactionDestinationEntry, 0),
            6: ('num_inputs', p.UVarintType, 0),
            7: ('mixin', p.UVarintType, 0),
            8: ('fee', p.UVarintType, 0),
            9: ('account', p.UVarintType, 0),
            10: ('minor_indices', p.UVarintType, p.FLAG_REPEATED),
            11: ('rsig_data', MoneroTransactionRsigData, 0),
            12: ('integrated_indices', p.UVarintType, p.FLAG_REPEATED),
            13: ('client_version', p.UVarintType, 0),
            14: ('hard_fork', p.UVarintType, 0),
            15: ('monero_version', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .EosTxHeader import EosTxHeader

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class EosSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 602

    def __init__(
        self,
        address_n: List[int] = None,
        chain_id: bytes = None,
        header: EosTxHeader = None,
        num_actions: int = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.chain_id = chain_id
        self.header = header
        self.num_actions = num_actions

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            2: ('chain_id', p.BytesType, 0),
            3: ('header', EosTxHeader, 0),
            4: ('num_actions', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroMultisigKLRki(p.MessageType):

    def __init__(
        self,
        K: bytes = None,
        L: bytes = None,
        R: bytes = None,
        ki: bytes = None,
    ) -> None:
        self.K = K
        self.L = L
        self.R = R
        self.ki = ki

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('K', p.BytesType, 0),
            2: ('L', p.BytesType, 0),
            3: ('R', p.BytesType, 0),
            4: ('ki', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class StellarSetOptionsOp(p.MessageType):
    MESSAGE_WIRE_TYPE = 215

    def __init__(
        self,
        source_account: str = None,
        inflation_destination_account: str = None,
        clear_flags: int = None,
        set_flags: int = None,
        master_weight: int = None,
        low_threshold: int = None,
        medium_threshold: int = None,
        high_threshold: int = None,
        home_domain: str = None,
        signer_type: int = None,
        signer_key: bytes = None,
        signer_weight: int = None,
    ) -> None:
        self.source_account = source_account
        self.inflation_destination_account = inflation_destination_account
        self.clear_flags = clear_flags
        self.set_flags = set_flags
        self.master_weight = master_weight
        self.low_threshold = low_threshold
        self.medium_threshold = medium_threshold
        self.high_threshold = high_threshold
        self.home_domain = home_domain
        self.signer_type = signer_type
        self.signer_key = signer_key
        self.signer_weight = signer_weight

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('source_account', p.UnicodeType, 0),
            2: ('inflation_destination_account', p.UnicodeType, 0),
            3: ('clear_flags', p.UVarintType, 0),
            4: ('set_flags', p.UVarintType, 0),
            5: ('master_weight', p.UVarintType, 0),
            6: ('low_threshold', p.UVarintType, 0),
            7: ('medium_threshold', p.UVarintType, 0),
            8: ('high_threshold', p.UVarintType, 0),
            9: ('home_domain', p.UnicodeType, 0),
            10: ('signer_type', p.UVarintType, 0),
            11: ('signer_key', p.BytesType, 0),
            12: ('signer_weight', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .TezosBallotOp import TezosBallotOp
from .TezosDelegationOp import TezosDelegationOp
from .TezosOriginationOp import TezosOriginationOp
from .TezosProposalOp import TezosProposalOp
from .TezosRevealOp import TezosRevealOp
from .TezosTransactionOp import TezosTransactionOp

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class TezosSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 152

    def __init__(
        self,
        address_n: List[int] = None,
        branch: bytes = None,
        reveal: TezosRevealOp = None,
        transaction: TezosTransactionOp = None,
        origination: TezosOriginationOp = None,
        delegation: TezosDelegationOp = None,
        proposal: TezosProposalOp = None,
        ballot: TezosBallotOp = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.branch = branch
        self.reveal = reveal
        self.transaction = transaction
        self.origination = origination
        self.delegation = delegation
        self.proposal = proposal
        self.ballot = ballot

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            2: ('branch', p.BytesType, 0),
            3: ('reveal', TezosRevealOp, 0),
            4: ('transaction', TezosTransactionOp, 0),
            5: ('origination', TezosOriginationOp, 0),
            6: ('delegation', TezosDelegationOp, 0),
            7: ('proposal', TezosProposalOp, 0),
            8: ('ballot', TezosBallotOp, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .EosAuthorization import EosAuthorization

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class EosActionUpdateAuth(p.MessageType):

    def __init__(
        self,
        account: int = None,
        permission: int = None,
        parent: int = None,
        auth: EosAuthorization = None,
    ) -> None:
        self.account = account
        self.permission = permission
        self.parent = parent
        self.auth = auth

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('account', p.UVarintType, 0),
            2: ('permission', p.UVarintType, 0),
            3: ('parent', p.UVarintType, 0),
            4: ('auth', EosAuthorization, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .TxInputType import TxInputType
from .TxOutputBinType import TxOutputBinType
from .TxOutputType import TxOutputType

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class TransactionType(p.MessageType):

    def __init__(
        self,
        version: int = None,
        inputs: List[TxInputType] = None,
        bin_outputs: List[TxOutputBinType] = None,
        lock_time: int = None,
        outputs: List[TxOutputType] = None,
        inputs_cnt: int = None,
        outputs_cnt: int = None,
        extra_data: bytes = None,
        extra_data_len: int = None,
        expiry: int = None,
        overwintered: bool = None,
        version_group_id: int = None,
        timestamp: int = None,
        branch_id: int = None,
    ) -> None:
        self.version = version
        self.inputs = inputs if inputs is not None else []
        self.bin_outputs = bin_outputs if bin_outputs is not None else []
        self.lock_time = lock_time
        self.outputs = outputs if outputs is not None else []
        self.inputs_cnt = inputs_cnt
        self.outputs_cnt = outputs_cnt
        self.extra_data = extra_data
        self.extra_data_len = extra_data_len
        self.expiry = expiry
        self.overwintered = overwintered
        self.version_group_id = version_group_id
        self.timestamp = timestamp
        self.branch_id = branch_id

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('version', p.UVarintType, 0),
            2: ('inputs', TxInputType, p.FLAG_REPEATED),
            3: ('bin_outputs', TxOutputBinType, p.FLAG_REPEATED),
            4: ('lock_time', p.UVarintType, 0),
            5: ('outputs', TxOutputType, p.FLAG_REPEATED),
            6: ('inputs_cnt', p.UVarintType, 0),
            7: ('outputs_cnt', p.UVarintType, 0),
            8: ('extra_data', p.BytesType, 0),
            9: ('extra_data_len', p.UVarintType, 0),
            10: ('expiry', p.UVarintType, 0),
            11: ('overwintered', p.BoolType, 0),
            12: ('version_group_id', p.UVarintType, 0),
            13: ('timestamp', p.UVarintType, 0),
            14: ('branch_id', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeTezosBallotType = Literal[0, 1, 2]
    except ImportError:
        pass


class TezosBallotOp(p.MessageType):

    def __init__(
        self,
        source: bytes = None,
        period: int = None,
        proposal: bytes = None,
        ballot: EnumTypeTezosBallotType = None,
    ) -> None:
        self.source = source
        self.period = period
        self.proposal = proposal
        self.ballot = ballot

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('source', p.BytesType, 0),
            2: ('period', p.UVarintType, 0),
            3: ('proposal', p.BytesType, 0),
            4: ('ballot', p.EnumType("TezosBallotType", (0, 1, 2)), 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class NEMMosaic(p.MessageType):

    def __init__(
        self,
        namespace: str = None,
        mosaic: str = None,
        quantity: int = None,
    ) -> None:
        self.namespace = namespace
        self.mosaic = mosaic
        self.quantity = quantity

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('namespace', p.UnicodeType, 0),
            2: ('mosaic', p.UnicodeType, 0),
            3: ('quantity', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class HDNodeType(p.MessageType):

    def __init__(
        self,
        depth: int = None,
        fingerprint: int = None,
        child_num: int = None,
        chain_code: bytes = None,
        private_key: bytes = None,
        public_key: bytes = None,
    ) -> None:
        self.depth = depth
        self.fingerprint = fingerprint
        self.child_num = child_num
        self.chain_code = chain_code
        self.private_key = private_key
        self.public_key = public_key

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('depth', p.UVarintType, 0),  # required
            2: ('fingerprint', p.UVarintType, 0),  # required
            3: ('child_num', p.UVarintType, 0),  # required
            4: ('chain_code', p.BytesType, 0),  # required
            5: ('private_key', p.BytesType, 0),
            6: ('public_key', p.BytesType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import functools
import sys

import click

from .. import exceptions
from ..client import TrezorClient
from ..transport import get_transport
from ..ui import ClickUI


class ChoiceType(click.Choice):
    def __init__(self, typemap):
        super().__init__(typemap.keys())
        self.typemap = typemap

    def convert(self, value, param, ctx):
        if value in self.typemap.values():
            return value
        value = super().convert(value, param, ctx)
        return self.typemap[value]


class TrezorConnection:
    def __init__(self, path, session_id, passphrase_on_host):
        self.path = path
        self.session_id = session_id
        self.passphrase_on_host = passphrase_on_host

    def get_transport(self):
        try:
            # look for transport without prefix search
            return get_transport(self.path, prefix_search=False)
        except Exception:
            # most likely not found. try again below.
            pass

        # look for transport with prefix search
        # if this fails, we want the exception to bubble up to the caller
        return get_transport(self.path, prefix_search=True)

    def get_ui(self):
        return ClickUI(passphrase_on_host=self.passphrase_on_host)

    def get_client(self):
        transport = self.get_transport()
        ui = self.get_ui()
        return TrezorClient(transport, ui=ui, session_id=self.session_id)


def with_client(func):
    @click.pass_obj
    @functools.wraps(func)
    def trezorctl_command_with_client(obj, *args, **kwargs):
        try:
            client = obj.get_client()
        except Exception:
            click.echo("Failed to find a Trezor device.")
            if obj.path is not None:
                click.echo("Using path: {}".format(obj.path))
            sys.exit(1)

        try:
            return func(client, *args, **kwargs)
        except exceptions.Cancelled:
            click.echo("Action was cancelled.")
            sys.exit(1)
        except exceptions.TrezorException as e:
            raise click.ClickException(str(e)) from e

    return trezorctl_command_with_client



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroGetAddress(p.MessageType):
    MESSAGE_WIRE_TYPE = 540

    def __init__(
        self,
        address_n: List[int] = None,
        show_display: bool = None,
        network_type: int = None,
        account: int = None,
        minor: int = None,
        payment_id: bytes = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.show_display = show_display
        self.network_type = network_type
        self.account = account
        self.minor = minor
        self.payment_id = payment_id

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            2: ('show_display', p.BoolType, 0),
            3: ('network_type', p.UVarintType, 0),
            4: ('account', p.UVarintType, 0),
            5: ('minor', p.UVarintType, 0),
            6: ('payment_id', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

Strict = 0  # type: Literal[0]
PromptAlways = 1  # type: Literal[1]
PromptTemporarily = 2  # type: Literal[2]



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import click

from .. import cosi, tools
from . import with_client

PATH_HELP = "BIP-32 path, e.g. m/44'/0'/0'/0/0"


@click.group(name="cosi")
def cli():
    """CoSi (Cothority / collective signing) commands."""


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.argument("data")
@with_client
def commit(client, address, data):
    """Ask device to commit to CoSi signing."""
    address_n = tools.parse_path(address)
    return cosi.commit(client, address_n, bytes.fromhex(data))


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.argument("data")
@click.argument("global_commitment")
@click.argument("global_pubkey")
@with_client
def sign(client, address, data, global_commitment, global_pubkey):
    """Ask device to sign using CoSi."""
    address_n = tools.parse_path(address)
    return cosi.sign(
        client,
        address_n,
        bytes.fromhex(data),
        bytes.fromhex(global_commitment),
        bytes.fromhex(global_pubkey),
    )



#!/usr/bin/env python3
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
import hmac
import hashlib
import json
import os
from urllib.parse import urlparse

from trezorlib import misc, ui
from trezorlib.client import TrezorClient
from trezorlib.transport import get_transport
from trezorlib.tools import parse_path


# Return path by BIP-32
BIP32_PATH = parse_path("10016h/0")


# Deriving master key
def getMasterKey(client):
    bip32_path = BIP32_PATH
    ENC_KEY = 'Activate TREZOR Password Manager?'
    ENC_VALUE = bytes.fromhex('2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee')
    key = misc.encrypt_keyvalue(
        client,
        bip32_path,
        ENC_KEY,
        ENC_VALUE,
        True,
        True
    )
    return key.hex()


# Deriving file name and encryption key
def getFileEncKey(key):
    filekey, enckey = key[:len(key) // 2], key[len(key) // 2:]
    FILENAME_MESS = b'5f91add3fa1c3c76e90c90a3bd0999e2bd7833d06a483fe884ee60397aca277a'
    digest = hmac.new(str.encode(filekey), FILENAME_MESS, hashlib.sha256).hexdigest()
    filename = digest + '.pswd'
    return [filename, filekey, enckey]


# File level decryption and file reading
def decryptStorage(path, key):
    cipherkey = bytes.fromhex(key)
    with open(path, 'rb') as f:
        iv = f.read(12)
        tag = f.read(16)
        cipher = Cipher(algorithms.AES(cipherkey), modes.GCM(iv, tag), backend=default_backend())
        decryptor = cipher.decryptor()
        data = ''
        while True:
            block = f.read(16)
            # data are not authenticated yet
            if block:
                data = data + decryptor.update(block).decode()
            else:
                break
        # throws exception when the tag is wrong
        data = data + decryptor.finalize().decode()
    return json.loads(data)


def decryptEntryValue(nonce, val):
    cipherkey = bytes.fromhex(nonce)
    iv = val[:12]
    tag = val[12:28]
    cipher = Cipher(algorithms.AES(cipherkey), modes.GCM(iv, tag), backend=default_backend())
    decryptor = cipher.decryptor()
    data = ''
    inputData = val[28:]
    while True:
        block = inputData[:16]
        inputData = inputData[16:]
        if block:
            data = data + decryptor.update(block).decode()
        else:
            break
        # throws exception when the tag is wrong
    data = data + decryptor.finalize().decode()
    return json.loads(data)


# Decrypt give entry nonce
def getDecryptedNonce(client, entry):
    print()
    print('Waiting for Trezor input ...')
    print()
    if 'item' in entry:
        item = entry['item']
    else:
        item = entry['title']

    pr = urlparse(item)
    if pr.scheme and pr.netloc:
        item = pr.netloc

    ENC_KEY = 'Unlock %s for user %s?' % (item, entry['username'])
    ENC_VALUE = entry['nonce']
    decrypted_nonce = misc.decrypt_keyvalue(
        client,
        BIP32_PATH,
        ENC_KEY,
        bytes.fromhex(ENC_VALUE),
        False,
        True
    )
    return decrypted_nonce.hex()


# Pretty print of list
def printEntries(entries):
    print('Password entries')
    print('================')
    print()
    for k, v in entries.items():
        print('Entry id: #%s' % k)
        print('-------------')
        for kk, vv in v.items():
            if kk in ['nonce', 'safe_note', 'password']:
                continue  # skip these fields
            print('*', kk, ': ', vv)
        print()
    return


def main():
    try:
        transport = get_transport()
    except Exception as e:
        print(e)
        return

    client = TrezorClient(transport=transport, ui=ui.ClickUI())

    print()
    print('Confirm operation on Trezor')
    print()

    masterKey = getMasterKey(client)
    # print('master key:', masterKey)

    fileName = getFileEncKey(masterKey)[0]
    # print('file name:', fileName)

    home = os.path.expanduser('~')
    path = os.path.join(home, 'Dropbox', 'Apps', 'TREZOR Password Manager')
    # print('path to file:', path)

    encKey = getFileEncKey(masterKey)[2]
    # print('enckey:', encKey)

    full_path = os.path.join(path, fileName)
    parsed_json = decryptStorage(full_path, encKey)

    # list entries
    entries = parsed_json['entries']
    printEntries(entries)

    entry_id = input('Select entry number to decrypt: ')
    entry_id = str(entry_id)

    plain_nonce = getDecryptedNonce(client, entries[entry_id])

    pwdArr = entries[entry_id]['password']['data']
    pwdHex = ''.join([hex(x)[2:].zfill(2) for x in pwdArr])
    print('password: ', decryptEntryValue(plain_nonce, bytes.fromhex(pwdHex)))

    safeNoteArr = entries[entry_id]['safe_note']['data']
    safeNoteHex = ''.join([hex(x)[2:].zfill(2) for x in safeNoteArr])
    print('safe_note:', decryptEntryValue(plain_nonce, bytes.fromhex(safeNoteHex)))

    return


if __name__ == '__main__':
    main()



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

Bitcoin = 1  # type: Literal[1]
Bitcoin_like = 2  # type: Literal[2]
Binance = 3  # type: Literal[3]
Cardano = 4  # type: Literal[4]
Crypto = 5  # type: Literal[5]
EOS = 6  # type: Literal[6]
Ethereum = 7  # type: Literal[7]
Lisk = 8  # type: Literal[8]
Monero = 9  # type: Literal[9]
NEM = 10  # type: Literal[10]
Ripple = 11  # type: Literal[11]
Stellar = 12  # type: Literal[12]
Tezos = 13  # type: Literal[13]
U2F = 14  # type: Literal[14]
Shamir = 15  # type: Literal[15]
ShamirGroups = 16  # type: Literal[16]
PassphraseEntry = 17  # type: Literal[17]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .NEMAggregateModification import NEMAggregateModification
from .NEMImportanceTransfer import NEMImportanceTransfer
from .NEMMosaicCreation import NEMMosaicCreation
from .NEMMosaicSupplyChange import NEMMosaicSupplyChange
from .NEMProvisionNamespace import NEMProvisionNamespace
from .NEMTransactionCommon import NEMTransactionCommon
from .NEMTransfer import NEMTransfer

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class NEMSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 69

    def __init__(
        self,
        transaction: NEMTransactionCommon = None,
        multisig: NEMTransactionCommon = None,
        transfer: NEMTransfer = None,
        cosigning: bool = None,
        provision_namespace: NEMProvisionNamespace = None,
        mosaic_creation: NEMMosaicCreation = None,
        supply_change: NEMMosaicSupplyChange = None,
        aggregate_modification: NEMAggregateModification = None,
        importance_transfer: NEMImportanceTransfer = None,
    ) -> None:
        self.transaction = transaction
        self.multisig = multisig
        self.transfer = transfer
        self.cosigning = cosigning
        self.provision_namespace = provision_namespace
        self.mosaic_creation = mosaic_creation
        self.supply_change = supply_change
        self.aggregate_modification = aggregate_modification
        self.importance_transfer = importance_transfer

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('transaction', NEMTransactionCommon, 0),
            2: ('multisig', NEMTransactionCommon, 0),
            3: ('transfer', NEMTransfer, 0),
            4: ('cosigning', p.BoolType, 0),
            5: ('provision_namespace', NEMProvisionNamespace, 0),
            6: ('mosaic_creation', NEMMosaicCreation, 0),
            7: ('supply_change', NEMMosaicSupplyChange, 0),
            8: ('aggregate_modification', NEMAggregateModification, 0),
            9: ('importance_transfer', NEMImportanceTransfer, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import io
from typing import Tuple

from . import messages, protobuf

map_type_to_class = {}
map_class_to_type = {}


def build_map():
    for msg_name in dir(messages.MessageType):
        if msg_name.startswith("__"):
            continue

        try:
            msg_class = getattr(messages, msg_name)
        except AttributeError:
            raise ValueError(
                "Implementation of protobuf message '%s' is missing" % msg_name
            )

        if msg_class.MESSAGE_WIRE_TYPE != getattr(messages.MessageType, msg_name):
            raise ValueError(
                "Inconsistent wire type and MessageType record for '%s'" % msg_class
            )

        register_message(msg_class)


def register_message(msg_class):
    if msg_class.MESSAGE_WIRE_TYPE in map_type_to_class:
        raise Exception(
            "Message for wire type %s is already registered by %s"
            % (msg_class.MESSAGE_WIRE_TYPE, get_class(msg_class.MESSAGE_WIRE_TYPE))
        )

    map_class_to_type[msg_class] = msg_class.MESSAGE_WIRE_TYPE
    map_type_to_class[msg_class.MESSAGE_WIRE_TYPE] = msg_class


def get_type(msg):
    return map_class_to_type[msg.__class__]


def get_class(t):
    return map_type_to_class[t]


def encode(msg: protobuf.MessageType) -> Tuple[int, bytes]:
    message_type = msg.MESSAGE_WIRE_TYPE
    buf = io.BytesIO()
    protobuf.dump_message(buf, msg)
    return message_type, buf.getvalue()


def decode(message_type: int, message_bytes: bytes) -> protobuf.MessageType:
    cls = get_class(message_type)
    buf = io.BytesIO(message_bytes)
    return protobuf.load_message(buf, cls)


build_map()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeInputScriptType = Literal[0, 1, 2, 3, 4]
    except ImportError:
        pass


class GetPublicKey(p.MessageType):
    MESSAGE_WIRE_TYPE = 11

    def __init__(
        self,
        address_n: List[int] = None,
        ecdsa_curve_name: str = None,
        show_display: bool = None,
        coin_name: str = None,
        script_type: EnumTypeInputScriptType = None,
        ignore_xpub_magic: bool = None,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.ecdsa_curve_name = ecdsa_curve_name
        self.show_display = show_display
        self.coin_name = coin_name
        self.script_type = script_type
        self.ignore_xpub_magic = ignore_xpub_magic

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            2: ('ecdsa_curve_name', p.UnicodeType, 0),
            3: ('show_display', p.BoolType, 0),
            4: ('coin_name', p.UnicodeType, 0),  # default=Bitcoin
            5: ('script_type', p.EnumType("InputScriptType", (0, 1, 2, 3, 4)), 0),  # default=SPENDADDRESS
            6: ('ignore_xpub_magic', p.BoolType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import click

from .. import monero, tools
from . import with_client

PATH_HELP = "BIP-32 path, e.g. m/44'/128'/0'"


@click.group(name="monero")
def cli():
    """Monero commands."""


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.option("-d", "--show-display", is_flag=True)
@click.option(
    "-t", "--network-type", type=click.Choice(["0", "1", "2", "3"]), default="0"
)
@with_client
def get_address(client, address, show_display, network_type):
    """Get Monero address for specified path."""
    address_n = tools.parse_path(address)
    network_type = int(network_type)
    return monero.get_address(client, address_n, show_display, network_type)


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.option(
    "-t", "--network-type", type=click.Choice(["0", "1", "2", "3"]), default="0"
)
@with_client
def get_watch_key(client, address, network_type):
    """Get Monero watch key for specified path."""
    address_n = tools.parse_path(address)
    network_type = int(network_type)
    res = monero.get_watch_key(client, address_n, network_type)
    output = {"address": res.address.decode(), "watch_key": res.watch_key.hex()}
    return output



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from io import BytesIO
import logging

import pytest

from trezorlib import protobuf


class PrimitiveMessage(protobuf.MessageType):
    @classmethod
    def get_fields(cls):
        return {
            1: ("uvarint", protobuf.UVarintType, 0),
            2: ("svarint", protobuf.SVarintType, 0),
            3: ("bool", protobuf.BoolType, 0),
            4: ("bytes", protobuf.BytesType, 0),
            5: ("unicode", protobuf.UnicodeType, 0),
            6: ("enum", protobuf.EnumType("t", (0, 5, 25)), 0),
        }


class EnumMessageMoreValues(protobuf.MessageType):
    @classmethod
    def get_fields(cls):
        return {1: ("enum", protobuf.EnumType("t", (0, 1, 2, 3, 4, 5)), 0)}


class EnumMessageLessValues(protobuf.MessageType):
    @classmethod
    def get_fields(cls):
        return {1: ("enum", protobuf.EnumType("t", (0, 5)), 0)}


class RepeatedFields(protobuf.MessageType):
    @classmethod
    def get_fields(cls):
        return {
            1: ("uintlist", protobuf.UVarintType, protobuf.FLAG_REPEATED),
            2: ("enumlist", protobuf.EnumType("t", (0, 1)), protobuf.FLAG_REPEATED),
            3: ("strlist", protobuf.UnicodeType, protobuf.FLAG_REPEATED),
        }


def load_uvarint(buffer):
    reader = BytesIO(buffer)
    return protobuf.load_uvarint(reader)


def dump_uvarint(value):
    writer = BytesIO()
    protobuf.dump_uvarint(writer, value)
    return writer.getvalue()


def test_dump_uvarint():
    assert dump_uvarint(0) == b"\x00"
    assert dump_uvarint(1) == b"\x01"
    assert dump_uvarint(0xFF) == b"\xff\x01"
    assert dump_uvarint(123456) == b"\xc0\xc4\x07"

    with pytest.raises(ValueError):
        dump_uvarint(-1)


def test_load_uvarint():
    assert load_uvarint(b"\x00") == 0
    assert load_uvarint(b"\x01") == 1
    assert load_uvarint(b"\xff\x01") == 0xFF
    assert load_uvarint(b"\xc0\xc4\x07") == 123456
    assert load_uvarint(b"\x80\x80\x80\x80\x00") == 0


def test_broken_uvarint():
    with pytest.raises(IOError):
        load_uvarint(b"\x80\x80")


def test_sint_uint():
    """
    Protobuf interleaved signed encoding
    https://developers.google.com/protocol-buffers/docs/encoding#structure
    LSbit is sign, rest is shifted absolute value.
    Or, by example, you count like so: 0, -1, 1, -2, 2, -3 ...
    """
    assert protobuf.sint_to_uint(0) == 0
    assert protobuf.uint_to_sint(0) == 0

    assert protobuf.sint_to_uint(-1) == 1
    assert protobuf.sint_to_uint(1) == 2

    assert protobuf.uint_to_sint(1) == -1
    assert protobuf.uint_to_sint(2) == 1

    # roundtrip:
    assert protobuf.uint_to_sint(protobuf.sint_to_uint(1234567891011)) == 1234567891011
    assert protobuf.uint_to_sint(protobuf.sint_to_uint(-2 ** 32)) == -2 ** 32


def test_simple_message():
    msg = PrimitiveMessage(
        uvarint=12345678910,
        svarint=-12345678910,
        bool=True,
        bytes=b"\xDE\xAD\xCA\xFE",
        unicode="Příliš žluťoučký kůň úpěl ďábelské ódy 😊",
        enum=5,
    )

    buf = BytesIO()

    protobuf.dump_message(buf, msg)
    buf.seek(0)
    retr = protobuf.load_message(buf, PrimitiveMessage)

    assert msg == retr
    assert retr.uvarint == 12345678910
    assert retr.svarint == -12345678910
    assert retr.bool is True
    assert retr.bytes == b"\xDE\xAD\xCA\xFE"
    assert retr.unicode == "Příliš žluťoučký kůň úpěl ďábelské ódy 😊"
    assert retr.enum == 5


def test_validate_enum(caplog):
    caplog.set_level(logging.INFO)
    # round-trip of a valid value
    msg = EnumMessageMoreValues(enum=0)
    buf = BytesIO()
    protobuf.dump_message(buf, msg)
    buf.seek(0)
    retr = protobuf.load_message(buf, EnumMessageLessValues)
    assert retr.enum == msg.enum

    assert not caplog.records

    # dumping an invalid enum value fails
    msg.enum = 19
    buf.seek(0)
    protobuf.dump_message(buf, msg)

    assert len(caplog.records) == 1
    record = caplog.records.pop(0)
    assert record.levelname == "INFO"
    assert record.getMessage() == "Value 19 unknown for type t"

    msg.enum = 3
    buf.seek(0)
    protobuf.dump_message(buf, msg)
    buf.seek(0)
    protobuf.load_message(buf, EnumMessageLessValues)

    assert len(caplog.records) == 1
    record = caplog.records.pop(0)
    assert record.levelname == "INFO"
    assert record.getMessage() == "Value 3 unknown for type t"


def test_repeated():
    msg = RepeatedFields(
        uintlist=[1, 2, 3], enumlist=[0, 1, 0, 1], strlist=["hello", "world"]
    )
    buf = BytesIO()
    protobuf.dump_message(buf, msg)
    buf.seek(0)
    retr = protobuf.load_message(buf, RepeatedFields)

    assert retr == msg


def test_enum_in_repeated(caplog):
    caplog.set_level(logging.INFO)

    msg = RepeatedFields(enumlist=[0, 1, 2, 3])
    buf = BytesIO()
    protobuf.dump_message(buf, msg)
    assert len(caplog.records) == 2
    for record in caplog.records:
        assert record.levelname == "INFO"
        assert "unknown for type t" in record.getMessage()


def test_packed():
    values = [4, 44, 444]
    packed_values = b"".join(dump_uvarint(v) for v in values)
    field_id = 1 << 3 | 2  # field number 1, wire type 2
    field_len = len(packed_values)
    message_bytes = dump_uvarint(field_id) + dump_uvarint(field_len) + packed_values

    buf = BytesIO(message_bytes)
    msg = protobuf.load_message(buf, RepeatedFields)
    assert msg
    assert msg.uintlist == values
    assert not msg.enumlist
    assert not msg.strlist


def test_packed_enum():
    values = [0, 0, 0, 0]
    packed_values = b"".join(dump_uvarint(v) for v in values)
    field_id = 2 << 3 | 2  # field number 2, wire type 2
    field_len = len(packed_values)
    message_bytes = dump_uvarint(field_id) + dump_uvarint(field_len) + packed_values

    buf = BytesIO(message_bytes)
    msg = protobuf.load_message(buf, RepeatedFields)
    assert msg
    assert msg.enumlist == values
    assert not msg.uintlist
    assert not msg.strlist



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class WebAuthnCredential(p.MessageType):

    def __init__(
        self,
        index: int = None,
        id: bytes = None,
        rp_id: str = None,
        rp_name: str = None,
        user_id: bytes = None,
        user_name: str = None,
        user_display_name: str = None,
        creation_time: int = None,
        hmac_secret: bool = None,
        use_sign_count: bool = None,
        algorithm: int = None,
        curve: int = None,
    ) -> None:
        self.index = index
        self.id = id
        self.rp_id = rp_id
        self.rp_name = rp_name
        self.user_id = user_id
        self.user_name = user_name
        self.user_display_name = user_display_name
        self.creation_time = creation_time
        self.hmac_secret = hmac_secret
        self.use_sign_count = use_sign_count
        self.algorithm = algorithm
        self.curve = curve

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('index', p.UVarintType, 0),
            2: ('id', p.BytesType, 0),
            3: ('rp_id', p.UnicodeType, 0),
            4: ('rp_name', p.UnicodeType, 0),
            5: ('user_id', p.BytesType, 0),
            6: ('user_name', p.UnicodeType, 0),
            7: ('user_display_name', p.UnicodeType, 0),
            8: ('creation_time', p.UVarintType, 0),
            9: ('hmac_secret', p.BoolType, 0),
            10: ('use_sign_count', p.BoolType, 0),
            11: ('algorithm', p.SVarintType, 0),
            12: ('curve', p.SVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeCapability = Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]
        EnumTypeBackupType = Literal[0, 1, 2]
        EnumTypeSafetyCheckLevel = Literal[0, 1, 2]
    except ImportError:
        pass


class Features(p.MessageType):
    MESSAGE_WIRE_TYPE = 17

    def __init__(
        self,
        vendor: str = None,
        major_version: int = None,
        minor_version: int = None,
        patch_version: int = None,
        bootloader_mode: bool = None,
        device_id: str = None,
        pin_protection: bool = None,
        passphrase_protection: bool = None,
        language: str = None,
        label: str = None,
        initialized: bool = None,
        revision: bytes = None,
        bootloader_hash: bytes = None,
        imported: bool = None,
        unlocked: bool = None,
        firmware_present: bool = None,
        needs_backup: bool = None,
        flags: int = None,
        model: str = None,
        fw_major: int = None,
        fw_minor: int = None,
        fw_patch: int = None,
        fw_vendor: str = None,
        fw_vendor_keys: bytes = None,
        unfinished_backup: bool = None,
        no_backup: bool = None,
        recovery_mode: bool = None,
        capabilities: List[EnumTypeCapability] = None,
        backup_type: EnumTypeBackupType = None,
        sd_card_present: bool = None,
        sd_protection: bool = None,
        wipe_code_protection: bool = None,
        session_id: bytes = None,
        passphrase_always_on_device: bool = None,
        safety_checks: EnumTypeSafetyCheckLevel = None,
    ) -> None:
        self.vendor = vendor
        self.major_version = major_version
        self.minor_version = minor_version
        self.patch_version = patch_version
        self.bootloader_mode = bootloader_mode
        self.device_id = device_id
        self.pin_protection = pin_protection
        self.passphrase_protection = passphrase_protection
        self.language = language
        self.label = label
        self.initialized = initialized
        self.revision = revision
        self.bootloader_hash = bootloader_hash
        self.imported = imported
        self.unlocked = unlocked
        self.firmware_present = firmware_present
        self.needs_backup = needs_backup
        self.flags = flags
        self.model = model
        self.fw_major = fw_major
        self.fw_minor = fw_minor
        self.fw_patch = fw_patch
        self.fw_vendor = fw_vendor
        self.fw_vendor_keys = fw_vendor_keys
        self.unfinished_backup = unfinished_backup
        self.no_backup = no_backup
        self.recovery_mode = recovery_mode
        self.capabilities = capabilities if capabilities is not None else []
        self.backup_type = backup_type
        self.sd_card_present = sd_card_present
        self.sd_protection = sd_protection
        self.wipe_code_protection = wipe_code_protection
        self.session_id = session_id
        self.passphrase_always_on_device = passphrase_always_on_device
        self.safety_checks = safety_checks

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('vendor', p.UnicodeType, 0),
            2: ('major_version', p.UVarintType, 0),
            3: ('minor_version', p.UVarintType, 0),
            4: ('patch_version', p.UVarintType, 0),
            5: ('bootloader_mode', p.BoolType, 0),
            6: ('device_id', p.UnicodeType, 0),
            7: ('pin_protection', p.BoolType, 0),
            8: ('passphrase_protection', p.BoolType, 0),
            9: ('language', p.UnicodeType, 0),
            10: ('label', p.UnicodeType, 0),
            12: ('initialized', p.BoolType, 0),
            13: ('revision', p.BytesType, 0),
            14: ('bootloader_hash', p.BytesType, 0),
            15: ('imported', p.BoolType, 0),
            16: ('unlocked', p.BoolType, 0),
            18: ('firmware_present', p.BoolType, 0),
            19: ('needs_backup', p.BoolType, 0),
            20: ('flags', p.UVarintType, 0),
            21: ('model', p.UnicodeType, 0),
            22: ('fw_major', p.UVarintType, 0),
            23: ('fw_minor', p.UVarintType, 0),
            24: ('fw_patch', p.UVarintType, 0),
            25: ('fw_vendor', p.UnicodeType, 0),
            26: ('fw_vendor_keys', p.BytesType, 0),
            27: ('unfinished_backup', p.BoolType, 0),
            28: ('no_backup', p.BoolType, 0),
            29: ('recovery_mode', p.BoolType, 0),
            30: ('capabilities', p.EnumType("Capability", (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17)), p.FLAG_REPEATED),
            31: ('backup_type', p.EnumType("BackupType", (0, 1, 2)), 0),
            32: ('sd_card_present', p.BoolType, 0),
            33: ('sd_protection', p.BoolType, 0),
            34: ('wipe_code_protection', p.BoolType, 0),
            35: ('session_id', p.BytesType, 0),
            36: ('passphrase_always_on_device', p.BoolType, 0),
            37: ('safety_checks', p.EnumType("SafetyCheckLevel", (0, 1, 2)), 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class DebugMoneroDiagRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 546

    def __init__(
        self,
        ins: int = None,
        p1: int = None,
        p2: int = None,
        pd: List[int] = None,
        data1: bytes = None,
        data2: bytes = None,
    ) -> None:
        self.ins = ins
        self.p1 = p1
        self.p2 = p2
        self.pd = pd if pd is not None else []
        self.data1 = data1
        self.data2 = data2

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('ins', p.UVarintType, 0),
            2: ('p1', p.UVarintType, 0),
            3: ('p2', p.UVarintType, 0),
            4: ('pd', p.UVarintType, p.FLAG_REPEATED),
            5: ('data1', p.BytesType, 0),
            6: ('data2', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class TezosDelegationOp(p.MessageType):

    def __init__(
        self,
        source: bytes = None,
        fee: int = None,
        counter: int = None,
        gas_limit: int = None,
        storage_limit: int = None,
        delegate: bytes = None,
    ) -> None:
        self.source = source
        self.fee = fee
        self.counter = counter
        self.gas_limit = gas_limit
        self.storage_limit = storage_limit
        self.delegate = delegate

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            7: ('source', p.BytesType, 0),
            2: ('fee', p.UVarintType, 0),
            3: ('counter', p.UVarintType, 0),
            4: ('gas_limit', p.UVarintType, 0),
            5: ('storage_limit', p.UVarintType, 0),
            6: ('delegate', p.BytesType, 0),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

Other = 1  # type: Literal[1]
FeeOverThreshold = 2  # type: Literal[2]
ConfirmOutput = 3  # type: Literal[3]
ResetDevice = 4  # type: Literal[4]
ConfirmWord = 5  # type: Literal[5]
WipeDevice = 6  # type: Literal[6]
ProtectCall = 7  # type: Literal[7]
SignTx = 8  # type: Literal[8]
FirmwareCheck = 9  # type: Literal[9]
Address = 10  # type: Literal[10]
PublicKey = 11  # type: Literal[11]
MnemonicWordCount = 12  # type: Literal[12]
MnemonicInput = 13  # type: Literal[13]
_Deprecated_ButtonRequest_PassphraseType = 14  # type: Literal[14]
UnknownDerivationPath = 15  # type: Literal[15]
RecoveryHomepage = 16  # type: Literal[16]
Success = 17  # type: Literal[17]
Warning = 18  # type: Literal[18]
PassphraseEntry = 19  # type: Literal[19]
PinEntry = 20  # type: Literal[20]



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

UP = 0  # type: Literal[0]
DOWN = 1  # type: Literal[1]
LEFT = 2  # type: Literal[2]
RIGHT = 3  # type: Literal[3]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .CardanoTxCertificateType import CardanoTxCertificateType
from .CardanoTxInputType import CardanoTxInputType
from .CardanoTxOutputType import CardanoTxOutputType
from .CardanoTxWithdrawalType import CardanoTxWithdrawalType

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class CardanoSignTx(p.MessageType):
    MESSAGE_WIRE_TYPE = 303

    def __init__(
        self,
        inputs: List[CardanoTxInputType] = None,
        outputs: List[CardanoTxOutputType] = None,
        protocol_magic: int = None,
        fee: int = None,
        ttl: int = None,
        network_id: int = None,
        certificates: List[CardanoTxCertificateType] = None,
        withdrawals: List[CardanoTxWithdrawalType] = None,
        metadata: bytes = None,
    ) -> None:
        self.inputs = inputs if inputs is not None else []
        self.outputs = outputs if outputs is not None else []
        self.protocol_magic = protocol_magic
        self.fee = fee
        self.ttl = ttl
        self.network_id = network_id
        self.certificates = certificates if certificates is not None else []
        self.withdrawals = withdrawals if withdrawals is not None else []
        self.metadata = metadata

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('inputs', CardanoTxInputType, p.FLAG_REPEATED),
            2: ('outputs', CardanoTxOutputType, p.FLAG_REPEATED),
            5: ('protocol_magic', p.UVarintType, 0),
            6: ('fee', p.UVarintType, 0),
            7: ('ttl', p.UVarintType, 0),
            8: ('network_id', p.UVarintType, 0),
            9: ('certificates', CardanoTxCertificateType, p.FLAG_REPEATED),
            10: ('withdrawals', CardanoTxWithdrawalType, p.FLAG_REPEATED),
            11: ('metadata', p.BytesType, 0),
        }



import struct
import zlib
from typing import Sequence, Tuple

import attr

from . import firmware

try:
    from PIL import Image
except ImportError:
    Image = None


RGBPixel = Tuple[int, int, int]


def _compress(data: bytes) -> bytes:
    z = zlib.compressobj(level=9, wbits=-10)
    return z.compress(data) + z.flush()


def _decompress(data: bytes) -> bytes:
    return zlib.decompress(data, wbits=-10)


def _from_pil_rgb(pixels: Sequence[RGBPixel]) -> bytes:
    data = bytearray()
    for r, g, b in pixels:
        c = ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | ((b & 0xF8) >> 3)
        data += struct.pack(">H", c)
    return bytes(data)


def _to_rgb(data: bytes) -> bytes:
    res = bytearray()
    for i in range(0, len(data), 2):
        (c,) = struct.unpack(">H", data[i : i + 2])
        r = (c & 0xF800) >> 8
        g = (c & 0x07C0) >> 3
        b = (c & 0x001F) << 3
        res += bytes((r, g, b))
    return bytes(res)


def _from_pil_grayscale(pixels: Sequence[int]) -> bytes:
    data = bytearray()
    for i in range(0, len(pixels), 2):
        left, right = pixels[i], pixels[i + 1]
        c = (left & 0xF0) | ((right & 0xF0) >> 4)
        data += struct.pack(">B", c)
    return bytes(data)


def _to_grayscale(data: bytes) -> bytes:
    res = bytearray()
    for pixel in data:
        left = pixel & 0xF0
        right = (pixel & 0x0F) << 4
        res += bytes((left, right))
    return bytes(res)


@attr.s
class Toif:
    mode = attr.ib()  # type: firmware.ToifMode
    size = attr.ib()  # type: Tuple[int, int]
    data = attr.ib()  # type: bytes

    @data.validator
    def check_data_size(self, _, value):
        width, height = self.size
        if self.mode is firmware.ToifMode.grayscale:
            expected_size = width * height // 2
        else:
            expected_size = width * height * 2
        uncompressed = _decompress(self.data)
        if len(uncompressed) != expected_size:
            raise ValueError(
                "Uncompressed data is {} bytes, expected {}".format(
                    len(uncompressed), expected_size
                )
            )

    def to_image(self) -> "Image":
        if Image is None:
            raise RuntimeError(
                "PIL is not available. Please install via 'pip install Pillow'"
            )

        uncompressed = _decompress(self.data)

        if self.mode is firmware.ToifMode.grayscale:
            pil_mode = "L"
            raw_data = _to_grayscale(uncompressed)
        else:
            pil_mode = "RGB"
            raw_data = _to_rgb(uncompressed)

        return Image.frombuffer(pil_mode, self.size, raw_data, "raw", pil_mode, 0, 1)

    def to_bytes(self) -> bytes:
        width, height = self.size
        return firmware.Toif.build(
            dict(format=self.mode, width=width, height=height, data=self.data)
        )

    def save(self, filename: str) -> None:
        with open(filename, "wb") as out:
            out.write(self.to_bytes())


def from_bytes(data: bytes) -> Toif:
    parsed = firmware.Toif.parse(data)
    return Toif(parsed.format, (parsed.width, parsed.height), parsed.data)


def load(filename: str) -> Toif:
    with open(filename, "rb") as f:
        return from_bytes(f.read())


def from_image(image: "Image", background=(0, 0, 0, 255)) -> Toif:
    if Image is None:
        raise RuntimeError(
            "PIL is not available. Please install via 'pip install Pillow'"
        )

    if image.mode == "RGBA":
        background = Image.new("RGBA", image.size, background)
        blend = Image.alpha_composite(background, image)
        image = blend.convert("RGB")

    if image.mode == "L":
        toif_mode = firmware.ToifMode.grayscale
        toif_data = _from_pil_grayscale(image.getdata())
    elif image.mode == "RGB":
        toif_mode = firmware.ToifMode.full_color
        toif_data = _from_pil_rgb(image.getdata())
    else:
        raise ValueError("Unsupported image mode: {}".format(image.mode))

    data = _compress(toif_data)
    return Toif(toif_mode, image.size, data)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class NEMProvisionNamespace(p.MessageType):

    def __init__(
        self,
        namespace: str = None,
        parent: str = None,
        sink: str = None,
        fee: int = None,
    ) -> None:
        self.namespace = namespace
        self.parent = parent
        self.sink = sink
        self.fee = fee

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('namespace', p.UnicodeType, 0),
            2: ('parent', p.UnicodeType, 0),
            3: ('sink', p.UnicodeType, 0),
            4: ('fee', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

ImportanceTransfer_Activate = 1  # type: Literal[1]
ImportanceTransfer_Deactivate = 2  # type: Literal[2]



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

Initialize = 0  # type: Literal[0]
Ping = 1  # type: Literal[1]
Success = 2  # type: Literal[2]
Failure = 3  # type: Literal[3]
ChangePin = 4  # type: Literal[4]
WipeDevice = 5  # type: Literal[5]
GetEntropy = 9  # type: Literal[9]
Entropy = 10  # type: Literal[10]
LoadDevice = 13  # type: Literal[13]
ResetDevice = 14  # type: Literal[14]
Features = 17  # type: Literal[17]
PinMatrixRequest = 18  # type: Literal[18]
PinMatrixAck = 19  # type: Literal[19]
Cancel = 20  # type: Literal[20]
LockDevice = 24  # type: Literal[24]
ApplySettings = 25  # type: Literal[25]
ButtonRequest = 26  # type: Literal[26]
ButtonAck = 27  # type: Literal[27]
ApplyFlags = 28  # type: Literal[28]
BackupDevice = 34  # type: Literal[34]
EntropyRequest = 35  # type: Literal[35]
EntropyAck = 36  # type: Literal[36]
PassphraseRequest = 41  # type: Literal[41]
PassphraseAck = 42  # type: Literal[42]
RecoveryDevice = 45  # type: Literal[45]
WordRequest = 46  # type: Literal[46]
WordAck = 47  # type: Literal[47]
GetFeatures = 55  # type: Literal[55]
SetU2FCounter = 63  # type: Literal[63]
SdProtect = 79  # type: Literal[79]
GetNextU2FCounter = 80  # type: Literal[80]
NextU2FCounter = 81  # type: Literal[81]
ChangeWipeCode = 82  # type: Literal[82]
EndSession = 83  # type: Literal[83]
DoPreauthorized = 84  # type: Literal[84]
PreauthorizedRequest = 85  # type: Literal[85]
CancelAuthorization = 86  # type: Literal[86]
RebootToBootloader = 87  # type: Literal[87]
Deprecated_PassphraseStateRequest = 77  # type: Literal[77]
Deprecated_PassphraseStateAck = 78  # type: Literal[78]
FirmwareErase = 6  # type: Literal[6]
FirmwareUpload = 7  # type: Literal[7]
FirmwareRequest = 8  # type: Literal[8]
SelfTest = 32  # type: Literal[32]
GetPublicKey = 11  # type: Literal[11]
PublicKey = 12  # type: Literal[12]
SignTx = 15  # type: Literal[15]
TxRequest = 21  # type: Literal[21]
TxAck = 22  # type: Literal[22]
GetAddress = 29  # type: Literal[29]
Address = 30  # type: Literal[30]
SignMessage = 38  # type: Literal[38]
VerifyMessage = 39  # type: Literal[39]
MessageSignature = 40  # type: Literal[40]
GetOwnershipId = 43  # type: Literal[43]
OwnershipId = 44  # type: Literal[44]
GetOwnershipProof = 49  # type: Literal[49]
OwnershipProof = 50  # type: Literal[50]
AuthorizeCoinJoin = 51  # type: Literal[51]
CipherKeyValue = 23  # type: Literal[23]
CipheredKeyValue = 48  # type: Literal[48]
SignIdentity = 53  # type: Literal[53]
SignedIdentity = 54  # type: Literal[54]
GetECDHSessionKey = 61  # type: Literal[61]
ECDHSessionKey = 62  # type: Literal[62]
CosiCommit = 71  # type: Literal[71]
CosiCommitment = 72  # type: Literal[72]
CosiSign = 73  # type: Literal[73]
CosiSignature = 74  # type: Literal[74]
DebugLinkDecision = 100  # type: Literal[100]
DebugLinkGetState = 101  # type: Literal[101]
DebugLinkState = 102  # type: Literal[102]
DebugLinkStop = 103  # type: Literal[103]
DebugLinkLog = 104  # type: Literal[104]
DebugLinkMemoryRead = 110  # type: Literal[110]
DebugLinkMemory = 111  # type: Literal[111]
DebugLinkMemoryWrite = 112  # type: Literal[112]
DebugLinkFlashErase = 113  # type: Literal[113]
DebugLinkLayout = 9001  # type: Literal[9001]
DebugLinkReseedRandom = 9002  # type: Literal[9002]
DebugLinkRecordScreen = 9003  # type: Literal[9003]
DebugLinkShowText = 9004  # type: Literal[9004]
DebugLinkEraseSdCard = 9005  # type: Literal[9005]
DebugLinkWatchLayout = 9006  # type: Literal[9006]
EthereumGetPublicKey = 450  # type: Literal[450]
EthereumPublicKey = 451  # type: Literal[451]
EthereumGetAddress = 56  # type: Literal[56]
EthereumAddress = 57  # type: Literal[57]
EthereumSignTx = 58  # type: Literal[58]
EthereumTxRequest = 59  # type: Literal[59]
EthereumTxAck = 60  # type: Literal[60]
EthereumSignMessage = 64  # type: Literal[64]
EthereumVerifyMessage = 65  # type: Literal[65]
EthereumMessageSignature = 66  # type: Literal[66]
NEMGetAddress = 67  # type: Literal[67]
NEMAddress = 68  # type: Literal[68]
NEMSignTx = 69  # type: Literal[69]
NEMSignedTx = 70  # type: Literal[70]
NEMDecryptMessage = 75  # type: Literal[75]
NEMDecryptedMessage = 76  # type: Literal[76]
LiskGetAddress = 114  # type: Literal[114]
LiskAddress = 115  # type: Literal[115]
LiskSignTx = 116  # type: Literal[116]
LiskSignedTx = 117  # type: Literal[117]
LiskSignMessage = 118  # type: Literal[118]
LiskMessageSignature = 119  # type: Literal[119]
LiskVerifyMessage = 120  # type: Literal[120]
LiskGetPublicKey = 121  # type: Literal[121]
LiskPublicKey = 122  # type: Literal[122]
TezosGetAddress = 150  # type: Literal[150]
TezosAddress = 151  # type: Literal[151]
TezosSignTx = 152  # type: Literal[152]
TezosSignedTx = 153  # type: Literal[153]
TezosGetPublicKey = 154  # type: Literal[154]
TezosPublicKey = 155  # type: Literal[155]
StellarSignTx = 202  # type: Literal[202]
StellarTxOpRequest = 203  # type: Literal[203]
StellarGetAddress = 207  # type: Literal[207]
StellarAddress = 208  # type: Literal[208]
StellarCreateAccountOp = 210  # type: Literal[210]
StellarPaymentOp = 211  # type: Literal[211]
StellarPathPaymentOp = 212  # type: Literal[212]
StellarManageOfferOp = 213  # type: Literal[213]
StellarCreatePassiveOfferOp = 214  # type: Literal[214]
StellarSetOptionsOp = 215  # type: Literal[215]
StellarChangeTrustOp = 216  # type: Literal[216]
StellarAllowTrustOp = 217  # type: Literal[217]
StellarAccountMergeOp = 218  # type: Literal[218]
StellarManageDataOp = 220  # type: Literal[220]
StellarBumpSequenceOp = 221  # type: Literal[221]
StellarSignedTx = 230  # type: Literal[230]
CardanoSignTx = 303  # type: Literal[303]
CardanoGetPublicKey = 305  # type: Literal[305]
CardanoPublicKey = 306  # type: Literal[306]
CardanoGetAddress = 307  # type: Literal[307]
CardanoAddress = 308  # type: Literal[308]
CardanoSignedTx = 310  # type: Literal[310]
RippleGetAddress = 400  # type: Literal[400]
RippleAddress = 401  # type: Literal[401]
RippleSignTx = 402  # type: Literal[402]
RippleSignedTx = 403  # type: Literal[403]
MoneroTransactionInitRequest = 501  # type: Literal[501]
MoneroTransactionInitAck = 502  # type: Literal[502]
MoneroTransactionSetInputRequest = 503  # type: Literal[503]
MoneroTransactionSetInputAck = 504  # type: Literal[504]
MoneroTransactionInputsPermutationRequest = 505  # type: Literal[505]
MoneroTransactionInputsPermutationAck = 506  # type: Literal[506]
MoneroTransactionInputViniRequest = 507  # type: Literal[507]
MoneroTransactionInputViniAck = 508  # type: Literal[508]
MoneroTransactionAllInputsSetRequest = 509  # type: Literal[509]
MoneroTransactionAllInputsSetAck = 510  # type: Literal[510]
MoneroTransactionSetOutputRequest = 511  # type: Literal[511]
MoneroTransactionSetOutputAck = 512  # type: Literal[512]
MoneroTransactionAllOutSetRequest = 513  # type: Literal[513]
MoneroTransactionAllOutSetAck = 514  # type: Literal[514]
MoneroTransactionSignInputRequest = 515  # type: Literal[515]
MoneroTransactionSignInputAck = 516  # type: Literal[516]
MoneroTransactionFinalRequest = 517  # type: Literal[517]
MoneroTransactionFinalAck = 518  # type: Literal[518]
MoneroKeyImageExportInitRequest = 530  # type: Literal[530]
MoneroKeyImageExportInitAck = 531  # type: Literal[531]
MoneroKeyImageSyncStepRequest = 532  # type: Literal[532]
MoneroKeyImageSyncStepAck = 533  # type: Literal[533]
MoneroKeyImageSyncFinalRequest = 534  # type: Literal[534]
MoneroKeyImageSyncFinalAck = 535  # type: Literal[535]
MoneroGetAddress = 540  # type: Literal[540]
MoneroAddress = 541  # type: Literal[541]
MoneroGetWatchKey = 542  # type: Literal[542]
MoneroWatchKey = 543  # type: Literal[543]
DebugMoneroDiagRequest = 546  # type: Literal[546]
DebugMoneroDiagAck = 547  # type: Literal[547]
MoneroGetTxKeyRequest = 550  # type: Literal[550]
MoneroGetTxKeyAck = 551  # type: Literal[551]
MoneroLiveRefreshStartRequest = 552  # type: Literal[552]
MoneroLiveRefreshStartAck = 553  # type: Literal[553]
MoneroLiveRefreshStepRequest = 554  # type: Literal[554]
MoneroLiveRefreshStepAck = 555  # type: Literal[555]
MoneroLiveRefreshFinalRequest = 556  # type: Literal[556]
MoneroLiveRefreshFinalAck = 557  # type: Literal[557]
EosGetPublicKey = 600  # type: Literal[600]
EosPublicKey = 601  # type: Literal[601]
EosSignTx = 602  # type: Literal[602]
EosTxActionRequest = 603  # type: Literal[603]
EosTxActionAck = 604  # type: Literal[604]
EosSignedTx = 605  # type: Literal[605]
BinanceGetAddress = 700  # type: Literal[700]
BinanceAddress = 701  # type: Literal[701]
BinanceGetPublicKey = 702  # type: Literal[702]
BinancePublicKey = 703  # type: Literal[703]
BinanceSignTx = 704  # type: Literal[704]
BinanceTxRequest = 705  # type: Literal[705]
BinanceTransferMsg = 706  # type: Literal[706]
BinanceOrderMsg = 707  # type: Literal[707]
BinanceCancelMsg = 708  # type: Literal[708]
BinanceSignedTx = 709  # type: Literal[709]
WebAuthnListResidentCredentials = 800  # type: Literal[800]
WebAuthnCredentials = 801  # type: Literal[801]
WebAuthnAddResidentCredential = 802  # type: Literal[802]
WebAuthnRemoveResidentCredential = 803  # type: Literal[803]



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class EosActionVoteProducer(p.MessageType):

    def __init__(
        self,
        voter: int = None,
        proxy: int = None,
        producers: List[int] = None,
    ) -> None:
        self.voter = voter
        self.proxy = proxy
        self.producers = producers if producers is not None else []

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('voter', p.UVarintType, 0),
            2: ('proxy', p.UVarintType, 0),
            3: ('producers', p.UVarintType, p.FLAG_REPEATED),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionDestinationEntry import MoneroTransactionDestinationEntry
from .MoneroTransactionRsigData import MoneroTransactionRsigData

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class MoneroTransactionSetOutputRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 511

    def __init__(
        self,
        dst_entr: MoneroTransactionDestinationEntry = None,
        dst_entr_hmac: bytes = None,
        rsig_data: MoneroTransactionRsigData = None,
        is_offloaded_bp: bool = None,
    ) -> None:
        self.dst_entr = dst_entr
        self.dst_entr_hmac = dst_entr_hmac
        self.rsig_data = rsig_data
        self.is_offloaded_bp = is_offloaded_bp

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('dst_entr', MoneroTransactionDestinationEntry, 0),
            2: ('dst_entr_hmac', p.BytesType, 0),
            3: ('rsig_data', MoneroTransactionRsigData, 0),
            4: ('is_offloaded_bp', p.BoolType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .EosAsset import EosAsset

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class EosActionDelegate(p.MessageType):

    def __init__(
        self,
        sender: int = None,
        receiver: int = None,
        net_quantity: EosAsset = None,
        cpu_quantity: EosAsset = None,
        transfer: bool = None,
    ) -> None:
        self.sender = sender
        self.receiver = receiver
        self.net_quantity = net_quantity
        self.cpu_quantity = cpu_quantity
        self.transfer = transfer

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('sender', p.UVarintType, 0),
            2: ('receiver', p.UVarintType, 0),
            3: ('net_quantity', EosAsset, 0),
            4: ('cpu_quantity', EosAsset, 0),
            5: ('transfer', p.BoolType, 0),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import json

import click

from .. import lisk, tools
from . import with_client

PATH_HELP = "BIP-32 path, e.g. m/44'/134'/0'/0'"


@click.group(name="lisk")
def cli():
    """Lisk commands."""


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.option("-d", "--show-display", is_flag=True)
@with_client
def get_address(client, address, show_display):
    """Get Lisk address for specified path."""
    address_n = tools.parse_path(address)
    return lisk.get_address(client, address_n, show_display)


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.option("-d", "--show-display", is_flag=True)
@with_client
def get_public_key(client, address, show_display):
    """Get Lisk public key for specified path."""
    address_n = tools.parse_path(address)
    res = lisk.get_public_key(client, address_n, show_display)
    output = {"public_key": res.public_key.hex()}
    return output


@cli.command()
@click.argument("file", type=click.File("r"))
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.option("-f", "--file", "_ignore", is_flag=True, hidden=True, expose_value=False)
@with_client
def sign_tx(client, address, file):
    """Sign Lisk transaction."""
    address_n = tools.parse_path(address)
    transaction = lisk.sign_tx(client, address_n, json.load(file))

    payload = {"signature": transaction.signature.hex()}

    return payload


@cli.command()
@click.option("-n", "--address", required=True, help=PATH_HELP)
@click.argument("message")
@with_client
def sign_message(client, address, message):
    """Sign message with Lisk address."""
    address_n = tools.parse_path(address)
    res = lisk.sign_message(client, address_n, message)
    output = {
        "message": message,
        "public_key": res.public_key.hex(),
        "signature": res.signature.hex(),
    }
    return output


@cli.command()
@click.argument("pubkey")
@click.argument("signature")
@click.argument("message")
@with_client
def verify_message(client, pubkey, signature, message):
    """Verify message signed with Lisk address."""
    signature = bytes.fromhex(signature)
    pubkey = bytes.fromhex(pubkey)
    return lisk.verify_message(client, pubkey, signature, message)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .NEMMosaic import NEMMosaic

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class NEMTransfer(p.MessageType):

    def __init__(
        self,
        recipient: str = None,
        amount: int = None,
        payload: bytes = None,
        public_key: bytes = None,
        mosaics: List[NEMMosaic] = None,
    ) -> None:
        self.recipient = recipient
        self.amount = amount
        self.payload = payload
        self.public_key = public_key
        self.mosaics = mosaics if mosaics is not None else []

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('recipient', p.UnicodeType, 0),
            2: ('amount', p.UVarintType, 0),
            3: ('payload', p.BytesType, 0),
            4: ('public_key', p.BytesType, 0),
            5: ('mosaics', NEMMosaic, p.FLAG_REPEATED),
        }



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

from typing import List

from . import messages, tools
from .tools import expect

PROTOCOL_MAGICS = {"mainnet": 764824073, "testnet": 42}
NETWORK_IDS = {"mainnet": 1, "testnet": 0}

REQUIRED_FIELDS_TRANSACTION = ("inputs", "outputs")
REQUIRED_FIELDS_INPUT = ("path", "prev_hash", "prev_index")
REQUIRED_FIELDS_CERTIFICATE = ("path", "type")
REQUIRED_FIELDS_WITHDRAWAL = ("path", "amount")

INCOMPLETE_OUTPUT_ERROR_MESSAGE = "The output is missing some fields"

ADDRESS_TYPES = (
    messages.CardanoAddressType.BYRON,
    messages.CardanoAddressType.BASE,
    messages.CardanoAddressType.POINTER,
    messages.CardanoAddressType.ENTERPRISE,
    messages.CardanoAddressType.REWARD,
)


def create_address_parameters(
    address_type: messages.CardanoAddressType,
    address_n: List[int],
    address_n_staking: List[int] = None,
    staking_key_hash: bytes = None,
    block_index: int = None,
    tx_index: int = None,
    certificate_index: int = None,
) -> messages.CardanoAddressParametersType:
    certificate_pointer = None

    if address_type not in ADDRESS_TYPES:
        raise ValueError("Unknown address type")

    if address_type == messages.CardanoAddressType.POINTER:
        certificate_pointer = create_certificate_pointer(
            block_index, tx_index, certificate_index
        )

    return messages.CardanoAddressParametersType(
        address_type=address_type,
        address_n=address_n,
        address_n_staking=address_n_staking,
        staking_key_hash=staking_key_hash,
        certificate_pointer=certificate_pointer,
    )


def create_certificate_pointer(
    block_index: int, tx_index: int, certificate_index: int
) -> messages.CardanoBlockchainPointerType:
    if block_index is None or tx_index is None or certificate_index is None:
        raise ValueError("Invalid pointer parameters")

    return messages.CardanoBlockchainPointerType(
        block_index=block_index, tx_index=tx_index, certificate_index=certificate_index
    )


def create_input(input) -> messages.CardanoTxInputType:
    if not all(k in input for k in REQUIRED_FIELDS_INPUT):
        raise ValueError("The input is missing some fields")

    path = input["path"]

    return messages.CardanoTxInputType(
        address_n=tools.parse_path(path),
        prev_hash=bytes.fromhex(input["prev_hash"]),
        prev_index=input["prev_index"],
    )


def create_output(output) -> messages.CardanoTxOutputType:
    contains_address = "address" in output
    contains_address_type = "addressType" in output

    if "amount" not in output:
        raise ValueError(INCOMPLETE_OUTPUT_ERROR_MESSAGE)
    if not (contains_address or contains_address_type):
        raise ValueError(INCOMPLETE_OUTPUT_ERROR_MESSAGE)

    if contains_address:
        return messages.CardanoTxOutputType(
            address=output["address"], amount=int(output["amount"])
        )
    else:
        return _create_change_output(output)


def _create_change_output(output) -> messages.CardanoTxOutputType:
    if "path" not in output:
        raise ValueError(INCOMPLETE_OUTPUT_ERROR_MESSAGE)

    staking_key_hash_bytes = None
    if "stakingKeyHash" in output:
        staking_key_hash_bytes = bytes.fromhex(output.get("stakingKeyHash"))

    address_parameters = create_address_parameters(
        int(output["addressType"]),
        tools.parse_path(output["path"]),
        tools.parse_path(output.get("stakingPath")),
        staking_key_hash_bytes,
        output.get("blockIndex"),
        output.get("txIndex"),
        output.get("certificateIndex"),
    )

    return messages.CardanoTxOutputType(
        address_parameters=address_parameters, amount=int(output["amount"])
    )


def create_certificate(certificate) -> messages.CardanoTxCertificateType:
    if not all(k in certificate for k in REQUIRED_FIELDS_CERTIFICATE):
        raise ValueError("The certificate is missing some fields")

    path = certificate["path"]
    certificate_type = certificate["type"]

    if certificate_type == messages.CardanoCertificateType.STAKE_DELEGATION:
        if "pool" not in certificate:
            raise ValueError("The certificate is missing some fields")

        pool = certificate["pool"]
        return messages.CardanoTxCertificateType(
            type=certificate_type,
            path=tools.parse_path(path),
            pool=bytes.fromhex(pool),
        )
    elif (
        certificate_type == messages.CardanoCertificateType.STAKE_REGISTRATION
        or certificate_type == messages.CardanoCertificateType.STAKE_DEREGISTRATION
    ):
        return messages.CardanoTxCertificateType(
            type=certificate_type, path=tools.parse_path(path),
        )
    else:
        raise ValueError("Unknown certificate type")


def create_withdrawal(withdrawal) -> messages.CardanoTxWithdrawalType:
    if not all(k in withdrawal for k in REQUIRED_FIELDS_WITHDRAWAL):
        raise ValueError("Withdrawal is missing some fields")

    path = withdrawal["path"]
    return messages.CardanoTxWithdrawalType(
        path=tools.parse_path(path), amount=int(withdrawal["amount"]),
    )


# ====== Client functions ====== #


@expect(messages.CardanoAddress, field="address")
def get_address(
    client,
    address_parameters: messages.CardanoAddressParametersType,
    protocol_magic: int = PROTOCOL_MAGICS["mainnet"],
    network_id: int = NETWORK_IDS["mainnet"],
    show_display: bool = False,
) -> messages.CardanoAddress:
    return client.call(
        messages.CardanoGetAddress(
            address_parameters=address_parameters,
            protocol_magic=protocol_magic,
            network_id=network_id,
            show_display=show_display,
        )
    )


@expect(messages.CardanoPublicKey)
def get_public_key(client, address_n: List[int]) -> messages.CardanoPublicKey:
    return client.call(messages.CardanoGetPublicKey(address_n=address_n))


@expect(messages.CardanoSignedTx)
def sign_tx(
    client,
    inputs: List[messages.CardanoTxInputType],
    outputs: List[messages.CardanoTxOutputType],
    fee: int,
    ttl: int,
    certificates: List[messages.CardanoTxCertificateType] = (),
    withdrawals: List[messages.CardanoTxWithdrawalType] = (),
    metadata: bytes = None,
    protocol_magic: int = PROTOCOL_MAGICS["mainnet"],
    network_id: int = NETWORK_IDS["mainnet"],
) -> messages.CardanoSignedTx:
    response = client.call(
        messages.CardanoSignTx(
            inputs=inputs,
            outputs=outputs,
            fee=fee,
            ttl=ttl,
            certificates=certificates,
            withdrawals=withdrawals,
            metadata=metadata,
            protocol_magic=protocol_magic,
            network_id=network_id,
        )
    )

    return response



# This file is part of the Trezor project.
#
# Copyright (C) 2012-2019 SatoshiLabs and contributors
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the License along with this library.
# If not, see <https://www.gnu.org/licenses/lgpl-3.0.html>.

import os

import click
from mnemonic import Mnemonic

from . import device
from .client import MAX_PIN_LENGTH, PASSPHRASE_ON_DEVICE
from .exceptions import Cancelled
from .messages import PinMatrixRequestType, WordRequestType

PIN_MATRIX_DESCRIPTION = """
Use the numeric keypad or lowercase letters to describe number positions.

The layout is:

    7 8 9        e r t
    4 5 6  -or-  d f g
    1 2 3        c v b
""".strip()

RECOVERY_MATRIX_DESCRIPTION = """
Use the numeric keypad to describe positions.
For the word list use only left and right keys.
Use backspace to correct an entry.

The keypad layout is:
    7 8 9     7 | 9
    4 5 6     4 | 6
    1 2 3     1 | 3
""".strip()

PIN_GENERIC = None
PIN_CURRENT = PinMatrixRequestType.Current
PIN_NEW = PinMatrixRequestType.NewFirst
PIN_CONFIRM = PinMatrixRequestType.NewSecond
WIPE_CODE_NEW = PinMatrixRequestType.WipeCodeFirst
WIPE_CODE_CONFIRM = PinMatrixRequestType.WipeCodeSecond


def echo(*args, **kwargs):
    return click.echo(*args, err=True, **kwargs)


def prompt(*args, **kwargs):
    return click.prompt(*args, err=True, **kwargs)


class ClickUI:
    def __init__(self, always_prompt=False, passphrase_on_host=False):
        self.pinmatrix_shown = False
        self.prompt_shown = False
        self.always_prompt = always_prompt
        self.passphrase_on_host = passphrase_on_host

    def button_request(self, code):
        if not self.prompt_shown:
            echo("Please confirm action on your Trezor device.")
        if not self.always_prompt:
            self.prompt_shown = True

    def get_pin(self, code=None):
        if code == PIN_CURRENT:
            desc = "current PIN"
        elif code == PIN_NEW:
            desc = "new PIN"
        elif code == PIN_CONFIRM:
            desc = "new PIN again"
        elif code == WIPE_CODE_NEW:
            desc = "new wipe code"
        elif code == WIPE_CODE_CONFIRM:
            desc = "new wipe code again"
        else:
            desc = "PIN"

        if not self.pinmatrix_shown:
            echo(PIN_MATRIX_DESCRIPTION)
            if not self.always_prompt:
                self.pinmatrix_shown = True

        while True:
            try:
                pin = prompt("Please enter {}".format(desc), hide_input=True)
            except click.Abort:
                raise Cancelled from None

            # translate letters to numbers if letters were used
            if all(d in "cvbdfgert" for d in pin):
                pin = pin.translate(str.maketrans("cvbdfgert", "123456789"))

            if any(d not in "123456789" for d in pin):
                echo(
                    "The value may only consist of digits 1 to 9 or letters cvbdfgert."
                )
            elif len(pin) > MAX_PIN_LENGTH:
                echo(
                    "The value must be at most {} digits in length.".format(
                        MAX_PIN_LENGTH
                    )
                )
            else:
                return pin

    def get_passphrase(self, available_on_device):
        if available_on_device and not self.passphrase_on_host:
            return PASSPHRASE_ON_DEVICE

        if os.getenv("PASSPHRASE") is not None:
            echo("Passphrase required. Using PASSPHRASE environment variable.")
            return os.getenv("PASSPHRASE")

        while True:
            try:
                passphrase = prompt(
                    "Passphrase required",
                    hide_input=True,
                    default="",
                    show_default=False,
                )
                second = prompt(
                    "Confirm your passphrase",
                    hide_input=True,
                    default="",
                    show_default=False,
                )
                if passphrase == second:
                    return passphrase
                else:
                    echo("Passphrase did not match. Please try again.")
            except click.Abort:
                raise Cancelled from None


def mnemonic_words(expand=False, language="english"):
    if expand:
        wordlist = Mnemonic(language).wordlist
    else:
        wordlist = set()

    def expand_word(word):
        if not expand:
            return word
        if word in wordlist:
            return word
        matches = [w for w in wordlist if w.startswith(word)]
        if len(matches) == 1:
            return matches[0]
        echo("Choose one of: " + ", ".join(matches))
        raise KeyError(word)

    def get_word(type):
        assert type == WordRequestType.Plain
        while True:
            try:
                word = prompt("Enter one word of mnemonic")
                return expand_word(word)
            except KeyError:
                pass
            except click.Abort:
                raise Cancelled from None

    return get_word


def matrix_words(type):
    while True:
        try:
            ch = click.getchar()
        except (KeyboardInterrupt, EOFError):
            raise Cancelled from None

        if ch in "\x04\x1b":
            # Ctrl+D, Esc
            raise Cancelled
        if ch in "\x08\x7f":
            # Backspace, Del
            return device.RECOVERY_BACK
        if type == WordRequestType.Matrix6 and ch in "147369":
            return ch
        if type == WordRequestType.Matrix9 and ch in "123456789":
            return ch



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .EosAuthorizationAccount import EosAuthorizationAccount
from .EosAuthorizationKey import EosAuthorizationKey
from .EosAuthorizationWait import EosAuthorizationWait

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class EosAuthorization(p.MessageType):

    def __init__(
        self,
        threshold: int = None,
        keys: List[EosAuthorizationKey] = None,
        accounts: List[EosAuthorizationAccount] = None,
        waits: List[EosAuthorizationWait] = None,
    ) -> None:
        self.threshold = threshold
        self.keys = keys if keys is not None else []
        self.accounts = accounts if accounts is not None else []
        self.waits = waits if waits is not None else []

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('threshold', p.UVarintType, 0),
            2: ('keys', EosAuthorizationKey, p.FLAG_REPEATED),
            3: ('accounts', EosAuthorizationAccount, p.FLAG_REPEATED),
            4: ('waits', EosAuthorizationWait, p.FLAG_REPEATED),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .CardanoBlockchainPointerType import CardanoBlockchainPointerType

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeCardanoAddressType = Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 14, 15]
    except ImportError:
        pass


class CardanoAddressParametersType(p.MessageType):

    def __init__(
        self,
        address_type: EnumTypeCardanoAddressType = None,
        address_n: List[int] = None,
        address_n_staking: List[int] = None,
        staking_key_hash: bytes = None,
        certificate_pointer: CardanoBlockchainPointerType = None,
    ) -> None:
        self.address_type = address_type
        self.address_n = address_n if address_n is not None else []
        self.address_n_staking = address_n_staking if address_n_staking is not None else []
        self.staking_key_hash = staking_key_hash
        self.certificate_pointer = certificate_pointer

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('address_type', p.EnumType("CardanoAddressType", (0, 1, 2, 3, 4, 5, 6, 7, 8, 14, 15)), 0),
            2: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            3: ('address_n_staking', p.UVarintType, p.FLAG_REPEATED),
            4: ('staking_key_hash', p.BytesType, 0),
            5: ('certificate_pointer', CardanoBlockchainPointerType, 0),
        }



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

OT_UNKNOWN = 0  # type: Literal[0]
MARKET = 1  # type: Literal[1]
LIMIT = 2  # type: Literal[2]
OT_RESERVED = 3  # type: Literal[3]



# Automatically generated by pb2py
# fmt: off
if False:
    from typing_extensions import Literal

UnexpectedMessage = 1  # type: Literal[1]
ButtonExpected = 2  # type: Literal[2]
DataError = 3  # type: Literal[3]
ActionCancelled = 4  # type: Literal[4]
PinExpected = 5  # type: Literal[5]
PinCancelled = 6  # type: Literal[6]
PinInvalid = 7  # type: Literal[7]
InvalidSignature = 8  # type: Literal[8]
ProcessError = 9  # type: Literal[9]
NotEnoughFunds = 10  # type: Literal[10]
NotInitialized = 11  # type: Literal[11]
PinMismatch = 12  # type: Literal[12]
WipeCodeMismatch = 13  # type: Literal[13]
FirmwareError = 99  # type: Literal[99]
